from typing import Any, Mapping, List, Union
from citylearn.agents.base import Agent
from citylearn.citylearn import CityLearnEnv
from citylearn.building import Building

class RBC(Agent):
    r"""Base rule based controller class.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """
    
    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

class HourRBC(RBC):
    r"""A hour-of-use rule-based controller.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    action_map: Union[Mapping[int, float], Mapping[str, Mapping[int, float]], List[Mapping[str, Mapping[int, float]]]], optional
        A 24-hour action map for each controlled device where the key is the hour between 1-24 and the value is the action.
        For storage systems, the value is negative for discharge and positive for charge where it ranges between [0, 1]. 
        Whereas, for cooling or heating devices, the value is positive for proportion of nominal power to make available and 
        ranges between [0, 1]. The action map can be parsed as a dictionary of hour keys mapped to action values 
        (:code:`Mapping[int, float]`). Alternatively, it can be parsed as a dictionary of devices 
        (:code:`Mapping[str, Mapping[int, float]]`) with their specific hour key to action value mapping 
        (:code:`Mapping[int, float]`). Finally, the action map can be defined for each agent especially in decentralized 
        setup as a list of device dictionary where each device dictionary (:code:`Mapping[str, Mapping[int, float]]`) 
        is for a specific decentralized (or centralized) agent. The HourRBC will return random actions if no map is provided.
    
    Other Parameters
    ----------------
    **kwargs: Any
        Other keyword arguments used to initialize super class.
    """
    
    
    def __init__(self, env: CityLearnEnv, action_map: Union[List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]] = None, **kwargs: Any):
        super().__init__(env, **kwargs)
        self.action_map = action_map

    @property
    def action_map(self) -> List[Mapping[str, Mapping[int, float]]]:
        return self.__action_map
    
    @action_map.setter
    def action_map(self, action_map: Union[List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]]):
        if isinstance(action_map, list):
            assert len(action_map) == len(self.action_dimension), f'List of action maps must have same length as number of agents: {len(self.action_dimension)}.'

            for i, (m, n) in enumerate(zip(action_map, self.action_names)):
                n = list(set(n))
                self.__verify_action_map(n, m, index=i)
        
        elif isinstance(action_map, dict):
            if isinstance(list(action_map.values())[0], dict):
                action_names = [a_ for a in self.action_names for a_ in a]
                action_names = list(set(action_names))
                self.__verify_action_map(action_names, action_map)
                action_map = [{n_: action_map[n_] for n_ in list(set(n))} for n in self.action_names]

            else:
                action_map = [{n_: action_map for n_ in list(set(n))} for n in self.action_names]
        
        else:
            pass

        self.__action_map = action_map

    def predict(self, observations: List[List[float]], deterministic: bool = None) -> List[List[float]]:
        """Provide actions for current time step.

        Parameters
        ----------
        observations: List[List[float]]
            Environment observations
        deterministic: bool, default: False
            Wether to return purely exploitatative deterministic actions.

        Returns
        -------
        actions: List[List[float]]
            Action values
        """

        actions = []

        if self.action_map is None:
            actions = super().predict(observations, deterministic=deterministic)
        
        else:
            for m, a, n, o in zip(self.action_map, self.action_names, self.observation_names, observations):
                hour_observation = o[n.index('hour')]
                hour = int(round(hour_observation))
                # Support both 0-23 and 1-24 hour encodings.
                hour_candidates = []

                for candidate in (hour, hour % 24, ((hour - 1) % 24) + 1):
                    if candidate not in hour_candidates:
                        hour_candidates.append(candidate)

                actions_ = []

                for a_ in a:
                    for candidate in hour_candidates:
                        hour_map = m[a_]

                        if candidate in hour_map:
                            actions_.append(hour_map[candidate])
                            break
                    else:
                        raise KeyError(f'Hour {hour_observation} not defined in action map for action {a_}.')
                
                actions.append(actions_)

            self.actions = actions
            self.next_time_step()
        
        return actions
    
    def __verify_action_map(self, action_names: List[str], action_map: Mapping[str, Mapping[int, float]], index: int = None):
        missing_actions = [a for a in action_names if a not in list(action_map.keys())]
        message = f'Undefined maps for actions: {missing_actions}'
        message += '.' if index is None else f' in building with index: {index}.'
        assert len(missing_actions) == 0, message

class BasicRBC(HourRBC):
    r"""A hour-of-use rule-based controller for heat-pump charged thermal energy storage systems that charges when COP is high.

    The actions are designed such that the agent charges the controlled storage system(s) by 9.1% of its maximum capacity every
    hour between 10:00 PM and 08:00 AM, and discharges 8.0% of its maximum capacity at every other hour. Cooling device is set
    to 40.0% of nominal power between between 10:00 PM and 08:00 AM and 80.0% at every other hour. Heating device is to 80.0% 
    of nominal power between between 10:00 PM and 08:00 AM and 40.0% at every other hour.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """
    
    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

    @HourRBC.action_map.setter
    def action_map(self, action_map: Union[List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]]):
        if action_map is None:
            action_map = {}
            action_names = [a_ for a in self.action_names for a_ in a]
            action_names = list(set(action_names))

            for n in action_names:
                action_map[n] = {}
                
                if 'storage' in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = -0.08
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.091
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = 0.8
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.4
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = 0.4
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.8
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_or_heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if hour < 7:
                            value = 0.4
                        
                        elif hour < 21:
                            value = -0.4

                        else:
                            value = 0.8

                        action_map[n][hour] = value
                
                else:
                    raise ValueError(f'Unknown action name: {n}')

        HourRBC.action_map.fset(self, action_map)

class OptimizedRBC(BasicRBC):
    r"""A hour-of-use rule-based controller that is an optimized version of :py:class:`citylearn.agents.rbc.BasicRBC`
    where control actions have been selected through a search grid.

    The actions are designed such that the agent discharges the controlled storage system(s) by 2.0% of its 
    maximum capacity every hour between 07:00 AM and 03:00 PM, discharges by 4.4% of its maximum capacity 
    between 04:00 PM and 06:00 PM, discharges by 2.4% of its maximum capacity between 07:00 PM and 10:00 PM, 
    charges by 3.4% of its maximum capacity between 11:00 PM to midnight and charges by 5.532% of its maximum 
    capacity at every other hour.

    Cooling device makes available 70.0% of its nominal power every hour between 07:00 AM and 03:00 PM, 60.0% between 
    04:00 PM and 06:00 PM, 80.0% between 07:00 PM and 10:00 PM, 40.0% between 11:00 PM to midnight and 20.0% other hour.

    Heating device makes available 30.0% of its nominal power every hour between 07:00 AM and 03:00 PM, 40.0% between 
    04:00 PM and 06:00 PM, 60.0% between 07:00 PM and 10:00 PM, 70.0% between 11:00 PM to midnight and 80.0% other hour.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """

    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

    @HourRBC.action_map.setter
    def action_map(self, action_map: Union[List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]]):
        if action_map is None:
            action_map = {}
            action_names = [a_ for a in self.action_names for a_ in a]
            action_names = list(set(action_names))

            for n in action_names:
                action_map[n] = {}
                
                if 'storage' in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 7 <= hour <= 15:
                            value = -0.02
                        elif 16 <= hour <= 18:
                            value = -0.044
                        elif 19 <= hour <= 22:
                            value = -0.024
                        elif 23 <= hour <= 24:
                            value = 0.034
                        elif 1 <= hour <= 6:
                            value = 0.05532
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 7 <= hour <= 15:
                            value = 0.7
                        elif 16 <= hour <= 18:
                            value = 0.6
                        elif 19 <= hour <= 22:
                            value = 0.8
                        elif 23 <= hour <= 24:
                            value = 0.4
                        elif 1 <= hour <= 6:
                            value = 0.2
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 7 <= hour <= 15:
                            value = 0.3
                        elif 16 <= hour <= 18:
                            value = 0.4
                        elif 19 <= hour <= 22:
                            value = 0.6
                        elif 23 <= hour <= 24:
                            value = 0.7
                        elif 1 <= hour <= 6:
                            value = 0.8
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_or_heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if hour < 7:
                            value = 0.4
                        
                        elif hour < 21:
                            value = -0.4

                        else:
                            value = 0.8

                        action_map[n][hour] = value
                
                else:
                    raise ValueError(f'Unknown action name: {n}')

        HourRBC.action_map.fset(self, action_map)

class BasicBatteryRBC(BasicRBC):
    r"""A hour-of-use rule-based controller that is designed to take advantage of solar generation for charging.

    The actions are optimized for electrical storage (battery) such that the agent charges the controlled
    storage system(s) by 11.0% of its maximum capacity every hour between 06:00 AM and 02:00 PM, 
    and discharges 6.7% of its maximum capacity at every other hour. Cooling device is set
    to 70.0% of nominal power between between 06:00 AM and 02:00 PM and 30.0% at every other hour. 
    Heating device is to 30.0% of nominal power between between 06:00 AM and 02:00 PM and 70.0% at every other hour

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    
    Other Parameters
    ----------------
    **kwargs: Any
        Other keyword arguments used to initialize super class.
    """

    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

    @HourRBC.action_map.setter
    def action_map(self, action_map: Union[List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]]):
        if action_map is None:
            action_map = {}
            action_names = [a_ for a in self.action_names for a_ in a]
            action_names = list(set(action_names))

            for n in action_names:
                action_map[n] = {}

                if 'storage' in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 6 <= hour <= 14:
                            value = 0.11
                        else:
                            value = -0.067

                        action_map[n][hour] = value

                elif n == 'cooling_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 6 <= hour <= 14:
                            value = 0.7
                        else:
                            value = 0.3

                        action_map[n][hour] = value

                elif n == 'heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 6 <= hour <= 14:
                            value = 0.3
                        else:
                            value = 0.7

                        action_map[n][hour] = value

                elif n == 'cooling_or_heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if hour < 7:
                            value = 0.4
                        
                        elif hour < 21:
                            value = -0.4

                        else:
                            value = 0.8

                        action_map[n][hour] = value
                
                else:
                    raise ValueError(f'Unknown action name: {n}')

        HourRBC.action_map.fset(self, action_map)

class BasicElectricVehicleRBC_ReferenceController(BasicRBC): #change the name
    r"""A rule-based controller that charges EVs to the maximum of the chargers and to the maximum of the battery upon connection.

    The EV battery is charged at the maximum available charging power whenever it is connected.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.

    Other Parameters
    ----------------
    **kwargs: Any
        Other keyword arguments used to initialize super class.
    """

    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        super().__init__(env, **kwargs)

    @HourRBC.action_map.setter
    def action_map(self, action_map: Union[
        List[Mapping[str, Mapping[int, float]]], Mapping[str, Mapping[int, float]], Mapping[int, float]]):
        if action_map is None:
            action_map = {}
            action_names = [a_ for a in self.action_names for a_ in a]
            action_names = list(set(action_names))

            for n in action_names:
                action_map[n] = {}

                if n == "electrical_storage":
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = -0.08
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.091
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = 0.8
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.4
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if 9 <= hour <= 21:
                            value = 0.4
                        elif (1 <= hour <= 8) or (22 <= hour <= 24):
                            value = 0.8
                        else:
                            value = 0.0

                        action_map[n][hour] = value

                elif n == 'cooling_or_heating_device':
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if hour < 7:
                            value = 0.4

                        elif hour < 21:
                            value = -0.4

                        else:
                            value = 0.8

                        action_map[n][hour] = value

                elif "electric_vehicle" in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        if hour < 7:
                            value = 0.4

                        elif hour < 10:
                            value = 1

                        elif hour < 15:
                            value = -1

                        elif hour < 20:
                            value = -0.6

                        else:
                            value = 0.8
                            
                        action_map[n][hour] = value

                elif "dhw_storage" in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        value = 1
                        action_map[n][hour] = value

                elif "washing_machine" in n:
                    for hour in Building.get_periodic_observation_metadata()['hour']:
                        value = 1
                        action_map[n][hour] = value        

                else:
                    raise ValueError(f'Unknown action name: {n}')

        HourRBC.action_map.fset(self, action_map)



from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.agents.rbc import BasicElectricVehicleRBC_ReferenceController as Agent
from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_three_phase_electrical_service_demo/schema.json"


def _zero_actions(env: CityLearnEnv):
    return [np.zeros(space.shape, dtype="float32") for space in env.action_space]


def test_charging_constraint_demo_rollout(tmp_path):
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        render_mode="during",
        render_directory=tmp_path / "SimulationData",
        episode_time_steps=48,
        random_seed=0,
    )

    try:
        controller = Agent(env)
        controller.learn(episodes=1, logging_level=1)
        kpis = controller.env.evaluate()
        assert not kpis.empty
    finally:
        env.close()


def test_charging_constraint_demo_zero_actions():
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=12,
        random_seed=0,
    )

    try:
        env.reset()
        while not env.terminated:
            env.step(_zero_actions(env))

        df = env.evaluate()
        finite = df["value"].dropna()
        assert not finite.empty
        assert np.isfinite(finite).all()
    finally:
        env.close()



import random
import numpy as np

# conditional imports
try:
    import torch
    from torch.distributions import Normal
    import torch.nn as nn
    import torch.nn.functional as F
except ImportError:
    raise Exception("This functionality requires you to install torch. You can install torch by : pip install torch torchvision, or for more detailed instructions please visit https://pytorch.org.")

class PolicyNetwork(nn.Module):
    def __init__(self, 
                 num_inputs, 
                 num_actions, 
                 action_space, 
                 action_scaling_coef, 
                 hidden_dim = [400,300],
                 init_w = 3e-3, 
                 log_std_min = -20, 
                 log_std_max = 2, 
                 epsilon = 1e-6):
        
        super(PolicyNetwork, self).__init__()

        self.log_std_min = log_std_min
        self.log_std_max = log_std_max
        self.epsilon = epsilon
        
        self.linear1 = nn.Linear(num_inputs, hidden_dim[0])
        self.linear2 = nn.Linear(hidden_dim[0], hidden_dim[1])

        self.mean_linear = nn.Linear(hidden_dim[1], num_actions)
        self.log_std_linear = nn.Linear(hidden_dim[1], num_actions)

        self.mean_linear.weight.data.uniform_(-init_w, init_w)
        self.mean_linear.bias.data.uniform_(-init_w, init_w)

        self.log_std_linear.weight.data.uniform_(-init_w, init_w)
        self.log_std_linear.bias.data.uniform_(-init_w, init_w)

        self.action_scale = torch.FloatTensor(
            action_scaling_coef * (action_space.high - action_space.low) / 2.)
        self.action_bias = torch.FloatTensor(
            action_scaling_coef * (action_space.high + action_space.low) / 2.)

    def forward(self, state):
        x = F.relu(self.linear1(state))
        x = F.relu(self.linear2(x))
        mean = self.mean_linear(x)
        log_std = self.log_std_linear(x)
        log_std = torch.clamp(log_std, min=self.log_std_min, max=self.log_std_max)
        return mean, log_std

    def sample(self, state):
        mean, log_std = self.forward(state)
        std = log_std.exp()
        normal = Normal(mean, std)
        x_t = normal.rsample()  # for reparameterization trick (mean + std * N(0,1))
        y_t = torch.tanh(x_t)
        action = y_t * self.action_scale + self.action_bias
        log_prob = normal.log_prob(x_t)
        # Enforcing Action Bound
        log_prob -= torch.log(self.action_scale * (1 - y_t.pow(2)) + self.epsilon)
        log_prob = log_prob.sum(1, keepdim=True)
        mean = torch.tanh(mean) * self.action_scale + self.action_bias
        return action, log_prob, mean

    def to(self, device):
        self.action_scale = self.action_scale.to(device)
        self.action_bias = self.action_bias.to(device)
        return super(PolicyNetwork, self).to(device)
    
class ReplayBuffer:
    def __init__(self, capacity):
        self.capacity = capacity
        self.buffer = []
        self.position = 0
    
    def push(self, state, action, reward, next_state, done):
        if len(self.buffer) < self.capacity:
            self.buffer.append(None)
        
        self.buffer[self.position] = (state, action, reward, next_state, done)
        self.position = (self.position + 1) % self.capacity
    
    def sample(self, batch_size):
        batch = random.sample(self.buffer, batch_size)
        state, action, reward, next_state, done = map(np.stack, zip(*batch))
        return state, action, reward, next_state, done
    
    def __len__(self):
        return len(self.buffer)
        
class RegressionBuffer:
    def __init__(self, capacity):
        self.capacity = capacity
        self.x = []
        self.y = []
        self.position = 0
    
    def push(self, variables, targets):
        if len(self.x) < self.capacity and len(self.x)==len(self.y):
            self.x.append(None)
            self.y.append(None)
        
        self.x[self.position] = variables
        self.y[self.position] = targets
        self.position = (self.position + 1) % self.capacity
    
    def __len__(self):
        return len(self.x)
    
class SoftQNetwork(nn.Module):
    def __init__(self, num_inputs, num_actions, hidden_size=[400,300], init_w=3e-3):
        super(SoftQNetwork, self).__init__()
        
        self.linear1 = nn.Linear(num_inputs + num_actions, hidden_size[0])
        self.linear2 = nn.Linear(hidden_size[0], hidden_size[1])
        self.linear3 = nn.Linear(hidden_size[1], 1)
        self.ln1 = nn.LayerNorm(hidden_size[0])
        self.ln2 = nn.LayerNorm(hidden_size[1])
        
        self.linear3.weight.data.uniform_(-init_w, init_w)
        self.linear3.bias.data.uniform_(-init_w, init_w)
        
    def forward(self, state, action):
        x = torch.cat([state, action], 1)
        x = self.ln1(F.relu(self.linear1(x)))
        x = self.ln2(F.relu(self.linear2(x)))
        x = self.linear3(x)
        return x


import random
from typing import Any, List, Mapping, Tuple, Union
import uuid
import numpy as np

class EpisodeTracker:
    """Class for keeping track of current episode time steps for reading observations from data files.

    An EpisodeTracker object is shared amongst the environment, buildings in environment and all descendant 
    building devices. The object however, should be updated at the environment level only so that its changes 
    propagate to all other evironment decscendants. `simulation_start_time_step` and `simulation_end_time_step`
    are useful to separate training data from test data in the same data file. There may be one or more episodes betweeen 
    `simulation_start_time_step` and `simulation_end_time_step` and their values should be defined in `schema` 
    or parsed to :py:class:`citylearn.citylearn.CityLearnEnv.__init__`. Both `simulation_start_time_step` and 
    `simulation_end_time_step` are used to select time series for building device and storage sizing as well as 
    action and observation space estimation in :py:class:`citylearn.buiLding.Building`.

    Parameters
    ----------
    simulation_start_time_step: int
        Time step to start reading from data files. 
    simulation_end_time_step: int
        Time step to end reading from data files.
    """
    
    def __init__(self, simulation_start_time_step: int, simulation_end_time_step: int):
        self.__episode = None
        self.__episode_start_time_step = None
        self.__episode_end_time_step = None
        self.__simulation_start_time_step = simulation_start_time_step
        self.__simulation_end_time_step = simulation_end_time_step
        self.reset_episode_index()

    @property
    def episode(self):
        """Current episode index"""

        return self.__episode
    
    @property
    def episode_time_steps(self):
        """Number of time steps in current episode split."""

        return (self.episode_end_time_step - self.episode_start_time_step) + 1
    
    @property
    def simulation_time_steps(self):
        """Number of time steps between `simulation_start_time_step` and `simulation_end_time_step`."""

        return (self.__simulation_end_time_step - self.__simulation_start_time_step) + 1
    
    @property
    def simulation_start_time_step(self):
        """Time step to start reading from data files."""

        return self.__simulation_start_time_step
    
    @property
    def simulation_end_time_step(self):
        """Time step to end reading from data files."""

        return self.__simulation_end_time_step
    
    @property
    def episode_start_time_step(self):
        """Start time step in current episode split."""

        return self.__episode_start_time_step
    
    @property
    def episode_end_time_step(self):
        """End time step in current episode split."""

        return self.__episode_end_time_step

    def next_episode(self, episode_time_steps: Union[int, List[Tuple[int, int]]], rolling_episode_split: bool, random_episode_split: bool, random_seed: int):
        """Advance to next episode and set `episode_start_time_step` and `episode_end_time_step` for reading data files.
        
        Parameters
        ----------
        episode_time_steps: Union[int, List[Tuple[int, int]]], optional
            If type is `int`, it is the number of time steps in an episode. If type is `List[Tuple[int, int]]]` is provided, it is a list of 
            episode start and end time steps between `simulation_start_time_step` and `simulation_end_time_step`. Defaults to (`simulation_end_time_step` 
        - `simulation_start_time_step`) + 1. Will ignore `rolling_episode_split` if `episode_splits` is of type `List[Tuple[int, int]]]`.
        rolling_episode_split: bool, default: False
            True if episode sequences are split such that each time step is a candidate for `episode_start_time_step` otherwise, False to split episodes 
            in steps of `episode_time_steps`.
        random_episode_split: bool, default: False
            True if episode splits are to be selected at random during training otherwise, False to select sequentially.
        """

        self.__episode += 1
        self.__next_episode_time_steps(
            episode_time_steps,
            rolling_episode_split,
            random_episode_split,
            random_seed,
        )
        
    def __next_episode_time_steps(self, episode_time_steps: Union[int, List[Tuple[int, int]]], rolling_episode_split: bool, random_episode_split: bool, random_seed: int):
        """Sets `episode_start_time_step` and `episode_end_time_step` for reading data files."""

        splits = None

        if isinstance(episode_time_steps, list):
            splits = episode_time_steps

        else:
            if episode_time_steps <= 0:
                raise ValueError(f'episode_time_steps must be >= 1, got {episode_time_steps}.')

            earliest_start_time_step = self.__simulation_start_time_step 
            latest_start_time_step = (self.__simulation_end_time_step + 1) - episode_time_steps

            if latest_start_time_step < earliest_start_time_step:
                raise ValueError(
                    f'episode_time_steps ({episode_time_steps}) exceeds available simulation window '
                    f'({self.simulation_time_steps}). Reduce episode_time_steps or adjust '
                    'simulation_start_time_step/simulation_end_time_step.'
                )
            
            if rolling_episode_split:
                start_time_steps = range(earliest_start_time_step, latest_start_time_step + 1)
            else:
                start_time_steps = range(earliest_start_time_step, latest_start_time_step + 1, episode_time_steps)

            end_time_steps = np.array(start_time_steps, dtype=int) + episode_time_steps - 1
            splits = np.array([start_time_steps, end_time_steps], dtype=int).T
            splits = splits.tolist()

        if len(splits) == 0:
            raise ValueError('No valid episode splits could be created from the provided episode_time_steps.')

        if random_episode_split:
            seed = int(random_seed*(self.episode + 1))
            nprs = np.random.RandomState(seed)
            ix = 0 if len(splits) == 1 else int(nprs.choice(len(splits)))

        else:
            ix = self.episode%len(splits)

        self.__episode_start_time_step, self.__episode_end_time_step = splits[ix]

    def reset_episode_index(self):
        """Resets episode index to -1 before any simulation."""

        self.__episode = -1

class Environment:
    """Base class for all `citylearn` classes that have a spatio-temporal dimension.

    Parameters
    ----------
    seconds_per_time_step: float, default: 3600.0
        Number of seconds in 1 `time_step` and must be set to >= 1.
    random_seed : int, optional
        Pseudorandom number generator seed for repeatable results.
    simulation_start_time_step: int, optional
        Time step to start reading from data files. Should be set at the :py:class:`citylearn.citylearn.CityLearnEnv` level so that it propagates to other descendant objects.
    simulation_end_time_step: int, optional
        Time step to end reading from data files. Should be set at the :py:class:`citylearn.citylearn.CityLearnEnv` level so that it propagates to other descendant objects.
    episode_tracker: EpisodeTracker, optional
        :py:class:`citylearn.base.EpisodeTracker` object used to keep track of current episode time steps for reading observations from data files.
    """

    DEFAULT_SECONDS_PER_TIME_STEP = 3600.0
    DEFAULT_RANDOM_SEED_RANGE = (0, 100_000_000)
    
    def __init__(self, seconds_per_time_step: float = None, random_seed: int = None, episode_tracker: EpisodeTracker = None, time_step_ratio: int = None):
        self.seconds_per_time_step = seconds_per_time_step
        self.__uid = uuid.uuid4().hex
        self.random_seed = random_seed
        self.__time_step = None
        self.episode_tracker = episode_tracker
        self.time_step_ratio = time_step_ratio

    @property
    def uid(self) -> str:
        r"""Unique environment ID."""

        return self.__uid
    
    @property
    def random_seed(self) -> int:
        """Pseudorandom number generator seed for repeatable results."""

        return self.__random_seed
    
    @property
    def episode_tracker(self) -> EpisodeTracker:
        """:py:class:`citylearn.base.EpisodeTracker` object used to keep track of 
        current episode time steps for reading observations from data files."""

        return self.__episode_tracker
    
    @property
    def time_step_ratio(self) -> int:
        """:py:class:`citylearn.base.EpisodeTracker` object used to keep track of 
        current episode time steps for reading observations from data files."""

        return self.__time_step_ratio

    @property
    def time_step(self) -> int:
        r"""Current environment time step."""

        return self.__time_step

    @property
    def seconds_per_time_step(self) -> float:
        r"""Number of seconds in 1 time step."""

        return self.__seconds_per_time_step
    
    @property
    def numpy_random_state(self) -> np.random.RandomState:
        """Nupy random state object."""

        return np.random.RandomState(self.random_seed)
    
    @random_seed.setter
    def random_seed(self, random_seed: int):
        random_seed = random.randint(*self.DEFAULT_RANDOM_SEED_RANGE) if random_seed is None else random_seed
        self.__random_seed = random_seed

    @seconds_per_time_step.setter
    def seconds_per_time_step(self, seconds_per_time_step: float):
        if seconds_per_time_step is None:
            self.seconds_per_time_step = self.DEFAULT_SECONDS_PER_TIME_STEP
        else:
            assert seconds_per_time_step >= 1, 'seconds_per_time_step >= 1'
            self.__seconds_per_time_step = seconds_per_time_step

    @episode_tracker.setter
    def episode_tracker(self, episode_tracker: EpisodeTracker):
        self.__episode_tracker = episode_tracker
    
    @time_step_ratio.setter
    def time_step_ratio(self, time_step_ratio: int):
        self.__time_step_ratio = time_step_ratio

    @time_step.setter
    def time_step(self, time_step: int):
        self.__time_step = time_step    

    def get_metadata(self) -> Mapping[str, Any]:
        """Returns general static information."""

        return {
            'uid': self.uid,
            'random_seed': self.random_seed,
            'simulation_time_steps': self.episode_tracker.simulation_time_steps,
            'seconds_per_time_step': self.seconds_per_time_step,
            'time_step_ratio': self.time_step_ratio,
        }

    def next_time_step(self):
        r"""Advance to next `time_step` value.

        Notes
        -----
        Override in subclass for custom implementation when advancing to next `time_step`.
        """

        self.__time_step += 1

    def reset(self):
        r"""Reset environment to initial state.

        Calls `reset_time_step`.

        Notes
        -----
        Override in subclass for custom implementation when reseting environment.
        """

        self.reset_time_step()

    def reset_time_step(self):
        r"""Reset `time_step` to initial state.

        Sets `time_step` to 0.
        """

        self.__time_step = 0



import json
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _run_fixed_episode(episode_time_steps: int = 12) -> pd.DataFrame:
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=episode_time_steps,
        random_seed=0,
    )

    try:
        env.reset()
        action_names = env.action_names[0]
        action = np.zeros(env.action_space[0].shape[0], dtype="float32")

        ev_indices = [i for i, name in enumerate(action_names) if name.startswith("electric_vehicle_storage_")]
        battery_indices = [i for i, name in enumerate(action_names) if "electrical_storage" in name]

        while not env.terminated:
            rollout_action = action.copy()
            if ev_indices:
                rollout_action[ev_indices] = 0.6
            if battery_indices:
                rollout_action[battery_indices] = 0.4

            _, _, terminated, truncated, _ = env.step([rollout_action])
            if terminated or truncated:
                break

        df = env.evaluate().sort_values(["level", "name", "cost_function"]).reset_index(drop=True)
        return df
    finally:
        env.close()


def test_loading_contract_shapes_and_counts_match_schema():
    schema = json.loads(SCHEMA.read_text(encoding="utf-8"))
    expected_buildings = [name for name, cfg in schema["buildings"].items() if cfg.get("include", False)]
    expected_evs = [name for name, cfg in schema.get("electric_vehicles_def", {}).items() if cfg.get("include", False)]

    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=False,
        episode_time_steps=6,
        random_seed=0,
    )

    try:
        env.reset()

        actual_buildings = [b.name for b in env.buildings]
        actual_evs = [ev.name for ev in env.electric_vehicles]

        assert actual_buildings == expected_buildings
        assert actual_evs == expected_evs

        assert len(env.observation_space) == len(env.buildings)
        assert len(env.action_space) == len(env.buildings)
        assert len(env.observation_names) == len(env.buildings)
        assert len(env.action_names) == len(env.buildings)

        for i, _ in enumerate(env.buildings):
            assert env.observation_space[i].shape[0] == len(env.observation_names[i])
            assert env.action_space[i].shape[0] == len(env.action_names[i])
    finally:
        env.close()


def test_kpi_results_are_stable_for_fixed_seed_and_actions():
    first = _run_fixed_episode(episode_time_steps=12)
    second = _run_fixed_episode(episode_time_steps=12)

    pd.testing.assert_frame_equal(first, second, check_exact=False, atol=1e-10, rtol=1e-10)



from collections import defaultdict
from copy import deepcopy
from enum import Enum
import hashlib
import importlib
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Mapping, Tuple, Union
from gymnasium import Env, spaces
import datetime
import numpy as np
import pandas as pd
import random
from citylearn.base import Environment, EpisodeTracker
from citylearn.building import Building, DynamicsBuilding
from citylearn.cost_function import CostFunction
from citylearn.data import CarbonIntensity, DataSet, ChargerSimulation, EnergySimulation, LogisticRegressionOccupantParameters, Pricing, WashingMachineSimulation, Weather
from citylearn.electric_vehicle import ElectricVehicle
from citylearn.energy_model import Battery, PV, WashingMachine
from citylearn.exporter import EpisodeExporter
from citylearn.internal.kpi import CityLearnKPIService
from citylearn.internal.loading import CityLearnLoadingService
from citylearn.internal.runtime import CityLearnRuntimeService
from citylearn.utilities import parse_bool
from citylearn.reward_function import (
    MultiBuildingRewardFunction,
    RewardFunction,
)
from citylearn.utilities import FileHandler

if TYPE_CHECKING:
    from citylearn.agents.base import Agent

LOGGER = logging.getLogger()
logging.getLogger('matplotlib.font_manager').disabled = True
logging.getLogger('matplotlib.pyplot').disabled = True

class EvaluationCondition(Enum):
    """Evaluation conditions.
    
    Used in `citylearn.CityLearnEnv.calculate` method.
    """

    # general (soft private)
    _DEFAULT = ''
    _STORAGE_SUFFIX = '_without_storage'
    _PARTIAL_LOAD_SUFFIX = '_and_partial_load'
    _PV_SUFFIX = '_and_pv'

    # Building type
    WITH_STORAGE_AND_PV = _DEFAULT
    WITHOUT_STORAGE_BUT_WITH_PV = _STORAGE_SUFFIX
    WITHOUT_STORAGE_AND_PV = WITHOUT_STORAGE_BUT_WITH_PV +_PV_SUFFIX

    # DynamicsBuilding type
    WITH_STORAGE_AND_PARTIAL_LOAD_AND_PV = WITH_STORAGE_AND_PV
    WITHOUT_STORAGE_BUT_WITH_PARTIAL_LOAD_AND_PV = WITHOUT_STORAGE_BUT_WITH_PV
    WITHOUT_STORAGE_AND_PARTIAL_LOAD_BUT_WITH_PV = WITHOUT_STORAGE_BUT_WITH_PARTIAL_LOAD_AND_PV + _PARTIAL_LOAD_SUFFIX
    WITHOUT_STORAGE_AND_PARTIAL_LOAD_AND_PV = WITHOUT_STORAGE_AND_PARTIAL_LOAD_BUT_WITH_PV + _PV_SUFFIX

class CityLearnEnv(Environment, Env):
    r"""CityLearn nvironment class.

    Parameters
    ----------
    schema: Union[str, Path, Mapping[str, Any]]
        Name of CityLearn data set, filepath to JSON representation or :code:`dict` object of a CityLearn schema.
        Call :py:meth:`citylearn.data.DataSet.get_names` for list of available CityLearn data sets.
    root_directory: Union[str, Path]
        Absolute path to directory that contains the data files including the schema.
    buildings: Union[List[Building], List[str], List[int]], optional
        Buildings to include in environment. If list of :code:`citylearn.building.Building` is provided, will override :code:`buildings` definition in schema.
        If list of :str: is provided will include only schema :code:`buildings` keys that are contained in provided list of :code:`str`.
        If list of :int: is provided will include only schema :code:`buildings` whose index is contained in provided list of :code:`int`.
    simulation_start_time_step: int, optional
        Time step to start reading data files contents.
    simulation_end_time_step: int, optional
        Time step to end reading from data files contents.
    episode_time_steps: Union[int, List[Tuple[int, int]]], optional
        If type is `int`, it is the number of time steps in an episode. If type is `List[Tuple[int, int]]]` is provided, 
        it is a list of episode start and end time steps between `simulation_start_time_step` and `simulation_end_time_step`. 
        Defaults to (`simulation_end_time_step` - `simulation_start_time_step`) + 1. Will ignore `rolling_episode_split` if `episode_splits` is of type `List[Tuple[int, int]]]`.
    rolling_episode_split: bool, default: False
        True if episode sequences are split such that each time step is a candidate for `episode_start_time_step` otherwise, False to split episodes in steps of `episode_time_steps`.
    random_episode_split: bool, default: False
        True if episode splits are to be selected at random during training otherwise, False to select sequentially.
    seconds_per_time_step: float
        Number of seconds in 1 `time_step` and must be set to >= 1.
    reward_function: Union[RewardFunction, str], optional
        Reward function class instance or path to function class e.g. 'citylearn.reward_function.IndependentSACReward'.
        If provided, will override :code:`reward_function` definition in schema.
    reward_function_kwargs: Mapping[str, Any], optional
        Parameters to be parsed to :py:attr:`reward_function` at intialization.
    central_agent: bool, optional
        Expect 1 central agent to control all buildings.
    shared_observations: List[str], optional
        Names of common observations across all buildings i.e. observations that have the same value irrespective of the building.
    active_observations: Union[List[str], List[List[str]]], optional
        List of observations to be made available in the buildings. Can be specified for all buildings in a :code:`List[str]` or for  
        each building independently in a :code:`List[List[str]]`. Will override the observations defined in the :code:`schema`.
    inactive_observations: Union[List[str], List[List[str]]], optional
        List of observations to be made unavailable in the buildings. Can be specified for all buildings in a :code:`List[str]` or for  
        each building independently in a :code:`List[List[str]]`. Will override the observations defined in the :code:`schema`.
    active_actions: Union[List[str], List[List[str]]], optional
        List of actions to be made available in the buildings. Can be specified for all buildings in a :code:`List[str]` or for  
        each building independently in a :code:`List[List[str]]`. Will override the actions defined in the :code:`schema`.
    inactive_actions: Union[List[str], List[List[str]]], optional
        List of actions to be made unavailable in the buildings. Can be specified for all buildings in a :code:`List[str]` or for  
        each building independently in a :code:`List[List[str]]`. Will override the actions defined in the :code:`schema`.
    simulate_power_outage: Union[bool, List[bool]]
        Whether to simulate power outages. Can be specified for all buildings as single :code:`bool` or for  
        each building independently in a :code:`List[bool]`. Will override power outage defined in the :code:`schema`.
    solar_generation: Union[bool, List[bool]]
        Wehther to allow solar generation. Can be specified for all buildings as single :code:`bool` or for  
        each building independently in a :code:`List[bool]`. Will override :code:`pv` defined in the :code:`schema`.
    random_seed: int, optional
        Pseudorandom number generator seed for repeatable results.

    Other Parameters
    ----------------
    render_directory: Union[str, Path], optional
        Base directory where rendering and export artifacts are stored. Relative paths are resolved from the project root.
    render_directory_name: str, optional
        Folder name created inside the project root for rendering and export artifacts when ``render_directory`` is not provided.
        Defaults to ``render_logs``.
    render_session_name: str, optional
        Name of the subfolder created under ``render_directory``/``render_directory_name`` for export artifacts. When omitted,
        a timestamp is used.
    render_mode: str, optional
        Rendering strategy. Accepted values are ``'none'`` (default), ``'during'`` for streaming exports each step, and
        ``'end'`` for exports performed at episode completion while still allowing manual snapshots via :meth:`render`.
    export_kpis_on_episode_end: bool, optional
        Whether to automatically export ``exported_kpis.csv`` when an episode terminates.
        If not provided, defaults to the effective rendering setting (enabled when rendering is enabled).
    **kwargs : dict
        Other keyword arguments used to initialize super classes.

    Notes
    -----
    Parameters passed to `citylearn.citylearn.CityLearnEnv.__init__` that are also defined in `schema` will override their `schema` definition.
    """

    DEFAULT_RENDER_START_DATE = datetime.date(2024, 1, 1)

    def __init__(self,
        schema: Union[str, Path, Mapping[str, Any]], root_directory: Union[str, Path] = None, buildings: Union[List[Building], List[str], List[int]] = None,
        electric_vehicles: Union[List[ElectricVehicle], List[str], List[int]] = None,
        simulation_start_time_step: int = None, simulation_end_time_step: int = None, episode_time_steps: Union[int, List[Tuple[int, int]]] = None, rolling_episode_split: bool = None,
        random_episode_split: bool = None, seconds_per_time_step: float = None, reward_function: Union[RewardFunction, str] = None, reward_function_kwargs: Mapping[str, Any] = None,
        central_agent: bool = None, shared_observations: List[str] = None, active_observations: Union[List[str], List[List[str]]] = None,
        inactive_observations: Union[List[str], List[List[str]]] = None, active_actions: Union[List[str], List[List[str]]] = None,
        inactive_actions: Union[List[str], List[List[str]]] = None, simulate_power_outage: bool = None, solar_generation: bool = None, random_seed: int = None, time_step_ratio: int = None,
        start_date: Union[str, datetime.date] = None, render_session_name: str = None, render_mode: str = 'none',
        export_kpis_on_episode_end: bool = None, **kwargs: Any
    ):
        render_directory = kwargs.pop('render_directory', None)
        render_directory_name = kwargs.pop('render_directory_name', 'render_logs')
        render_flag = kwargs.pop('render', None)
        kw_export_kpis_on_episode_end = kwargs.pop('export_kpis_on_episode_end', None)
        if kw_export_kpis_on_episode_end is not None and export_kpis_on_episode_end is None:
            export_kpis_on_episode_end = kw_export_kpis_on_episode_end
        debug_timing = kwargs.pop('debug_timing', None)
        check_observation_limits = kwargs.pop('check_observation_limits', None)
        metrics_log_interval = kwargs.pop('metrics_log_interval', None)
        kw_render_mode = kwargs.pop('render_mode', None)
        requested_render_mode = render_mode if kw_render_mode is None else kw_render_mode
        requested_render_mode = 'none' if requested_render_mode is None else str(requested_render_mode).lower()
        kw_render_session_name = kwargs.pop('render_session_name', None)
        if kw_render_session_name is not None:
            render_session_name = kw_render_session_name if render_session_name is None else render_session_name
        self.schema = schema
        self.community_market_enabled = False
        self.community_market_sell_ratio = 0.8
        self.community_market_grid_export_price = 0.0
        self._last_community_market_settlement = []
        self._community_market_settlement_history = []
        self._configure_community_market()
        schema_start_date = self.schema.get('start_date') if isinstance(self.schema, dict) else None
        schema_render_mode = self.schema.get('render_mode') if isinstance(self.schema, dict) else None
        schema_export_kpis = self.schema.get('export_kpis_on_episode_end') if isinstance(self.schema, dict) else None
        if schema_export_kpis is not None and export_kpis_on_episode_end is None:
            export_kpis_on_episode_end = parse_bool(
                schema_export_kpis,
                default=False,
                path='export_kpis_on_episode_end',
            )
        if schema_render_mode is not None:
            requested_render_mode = str(schema_render_mode).lower()
        if requested_render_mode not in {'none', 'during', 'end'}:
            raise ValueError("render_mode must be one of {'none', 'during', 'end'}.")
        self.render_mode = requested_render_mode
        self._buffer_render = False
        self._defer_render_flush = False
        self._render_buffer = defaultdict(list)
        self.debug_timing = parse_bool(
            self.schema.get('debug_timing', False) if debug_timing is None else debug_timing,
            default=False,
            path='debug_timing',
        )
        self.check_observation_limits = parse_bool(
            self.schema.get('check_observation_limits', False) if check_observation_limits is None else check_observation_limits,
            default=False,
            path='check_observation_limits',
        )
        self.metrics_log_interval = int(self.schema.get('metrics_log_interval', 0) if metrics_log_interval is None else metrics_log_interval)
        self._observations_cache: List[List[float]] = None
        self._observations_cache_time_step: int = -1
        self._render_start_date = self._parse_render_start_date(start_date if start_date is not None else schema_start_date)
        self.previous_month = None
        self.current_day = self._render_start_date.day
        self.year = self._render_start_date.year
        self._final_kpis_exported = False
        self.__rewards = None
        self.buildings = []
        self.random_seed = self.schema.get('random_seed', None) if random_seed is None else random_seed
        schema_render_session = self.schema.get('render_session_name') if isinstance(self.schema, dict) else None
        self.render_session_name = render_session_name if render_session_name is not None else schema_render_session
        if self.render_session_name is not None:
            self.render_session_name = str(self.render_session_name).strip()
            if self.render_session_name == '':
                self.render_session_name = None
            elif Path(self.render_session_name).is_absolute():
                raise ValueError('render_session_name must be a relative path. Use render_directory to choose an absolute location.')
            elif '..' in Path(self.render_session_name).parts:
                raise ValueError('render_session_name cannot contain parent directory references (“..”).')
        self._loading_service = CityLearnLoadingService(self)
        self._runtime_service = CityLearnRuntimeService(self)
        self._kpi_service = CityLearnKPIService(self)
        root_directory, buildings, electric_vehicles, episode_time_steps, rolling_episode_split, random_episode_split, \
            seconds_per_time_step, reward_function, central_agent, shared_observations, episode_tracker = self._load(
                deepcopy(self.schema),
                root_directory=root_directory,
                buildings=buildings,
                electric_vehicles=electric_vehicles,
                simulation_start_time_step=simulation_start_time_step,
                simulation_end_time_step=simulation_end_time_step,
                episode_time_steps=episode_time_steps,
                rolling_episode_split=rolling_episode_split,
                random_episode=random_episode_split,
                seconds_per_time_step=seconds_per_time_step,
                time_step_ratio=time_step_ratio,
                reward_function=reward_function,
                reward_function_kwargs=reward_function_kwargs,
                central_agent=central_agent,
                shared_observations=shared_observations,
                active_observations=active_observations,
                inactive_observations=inactive_observations,
                active_actions=active_actions,
                inactive_actions=inactive_actions,
                simulate_power_outage=simulate_power_outage,
                solar_generation=solar_generation,
                random_seed=self.random_seed,
            )
        self.root_directory = root_directory
        self.buildings = buildings
        self.electric_vehicles = electric_vehicles
        get_time_step_ratio = buildings[0].time_step_ratio if len(buildings) > 0 else 1.0
        self.time_step_ratio = get_time_step_ratio

        # now call super class initialization and set episode tracker now that buildings are set
        super().__init__(seconds_per_time_step=seconds_per_time_step, random_seed=self.random_seed, episode_tracker=episode_tracker, time_step_ratio=self.time_step_ratio)

        # set other class variables
        self.episode_time_steps = episode_time_steps
        self.rolling_episode_split = rolling_episode_split
        self.random_episode_split = random_episode_split
        self.central_agent = central_agent
        self.shared_observations = shared_observations

        # set reward function
        self.reward_function = reward_function
        self._refresh_action_cache()

        # rendering switch: schema['render'] overrides explicit flag, otherwise rely on render_mode defaults
        schema_render = self.schema.get('render', None) if isinstance(self.schema, dict) else None
        if schema_render is not None:
            render_enabled_flag = parse_bool(schema_render, default=False, path='render')
        elif render_flag is not None:
            render_enabled_flag = parse_bool(render_flag, default=False, path='render')
        else:
            render_enabled_flag = self.render_mode in {'during', 'end'}

        self.render_enabled = render_enabled_flag
        if export_kpis_on_episode_end is None:
            export_kpis_on_episode_end = self.render_enabled
        else:
            export_kpis_on_episode_end = parse_bool(
                export_kpis_on_episode_end,
                default=self.render_enabled,
                path='export_kpis_on_episode_end',
            )
        self.export_kpis_on_episode_end = export_kpis_on_episode_end

        # reset environment and initializes episode time steps
        self.reset()

        # reset episode tracker to start after initializing episode time steps during reset
        self.episode_tracker.reset_episode_index()

        # set reward metadata
        self.reward_function.env_metadata = self.get_metadata()

        # reward history tracker
        self.__episode_rewards = []

        # reward history tracker

        if self.root_directory is None:
            self.root_directory = os.path.dirname(os.path.abspath(__file__))

        project_root = Path(__file__).resolve().parents[1]
        render_directory_name = render_directory_name or 'render_logs'

        if render_directory is not None:
            render_root = Path(render_directory).expanduser()
            if not render_root.is_absolute():
                render_root = project_root / render_root
        else:
            render_root = project_root / render_directory_name

        self.render_output_root = render_root.expanduser().resolve()
        self._render_timestamp = None
        self._render_directory_path = None
        self._render_dir_initialized = False
        self.new_folder_path = None
        self._render_start_datetime = None
        self._episode_exporter = EpisodeExporter(self)

        if self.render_enabled:
            self._ensure_render_output_dir(ensure_exists=False)

    @property
    def render_start_date(self) -> datetime.date:
        """Date used as the origin for rendered timestamps."""

        return self._render_start_date

    @property
    def schema(self) -> Mapping[str, Any]:
        """`dict` object of CityLearn schema."""

        return self.__schema

    @property
    def render_enabled(self) -> bool:
        """Whether environment rendering/logging is enabled."""

        return getattr(self, '_CityLearnEnv__render_enabled', False)

    @property
    def export_kpis_on_episode_end(self) -> bool:
        """Whether KPIs are exported automatically when an episode terminates."""

        return getattr(self, '_CityLearnEnv__export_kpis_on_episode_end', False)

    @property
    def root_directory(self) -> Union[str, Path]:
        """Absolute path to directory that contains the data files including the schema."""

        return self.__root_directory

    @property
    def buildings(self) -> List[Building]:
        """Buildings in CityLearn environment."""

        return self.__buildings

    @property
    def electric_vehicles(self) -> List[ElectricVehicle]:
        """Electric Vehicles in CityLearn environment."""

        return self.__electric_vehicles

    @property
    def time_steps(self) -> int:
        """Number of time steps in current episode split."""

        return self.episode_tracker.episode_time_steps

    @property
    def episode_time_steps(self) -> Union[int, List[Tuple[int, int]]]:
        """If type is `int`, it is the number of time steps in an episode. If type is `List[Tuple[int, int]]]` is provided, it is a list of 
        episode start and end time steps between `simulation_start_time_step` and `simulation_end_time_step`. Defaults to (`simulation_end_time_step` 
        - `simulation_start_time_step`) + 1. Will ignore `rolling_episode_split` if `episode_splits` is of type `List[Tuple[int, int]]]`."""

        return self.__episode_time_steps

    @property
    def rolling_episode_split(self) -> bool:
        """True if episode sequences are split such that each time step is a candidate for `episode_start_time_step` otherwise, 
        False to split episodes in steps of `episode_time_steps`."""

        return self.__rolling_episode_split

    @property
    def random_episode_split(self) -> bool:
        """True if episode splits are to be selected at random during training otherwise, False to select sequentially."""

        return self.__random_episode_split

    @property
    def episode(self) -> int:
        """Current episode index."""

        return self.episode_tracker.episode

    @property
    def reward_function(self) -> RewardFunction:
        """Reward function class instance."""

        return self.__reward_function

    @property
    def rewards(self) -> List[List[float]]:
        """Reward time series"""

        return self.__rewards

    @property
    def episode_rewards(self) -> List[Mapping[str, Union[float, List[float]]]]:
        """Reward summary statistics for elapsed episodes."""

        return self.__episode_rewards

    @property
    def central_agent(self) -> bool:
        """Expect 1 central agent to control all buildings."""

        return self.__central_agent

    @property
    def shared_observations(self) -> List[str]:
        """Names of common observations across all buildings i.e. observations that have the same value irrespective of the building."""

        return self.__shared_observations

    @property
    def terminated(self) -> bool:
        """Check if simulation has reached completion."""

        return self.time_step >= self.time_steps - 1

    @property
    def truncated(self) -> bool:
        """Check if episode truncates due to a time limit or a reason that is not defined as part of the task MDP."""

        return False

    @property
    def observation_space(self) -> List[spaces.Box]:
        """Controller(s) observation spaces.

        Returns
        -------
        observation_space : List[spaces.Box]
            List of agent(s) observation spaces.
        
        Notes
        -----
        If `central_agent` is True, a list of 1 `spaces.Box` object is returned that contains all buildings' limits with the limits in the same order as `buildings`. 
        The `shared_observations` limits are only included in the first building's limits. If `central_agent` is False, a list of `space.Box` objects as
        many as `buildings` is returned in the same order as `buildings`.
        """

        if self.central_agent:
            low_limit = []
            high_limit = []
            shared_observations = []

            for i, b in enumerate(self.buildings):
                for l, h, s in zip(b.observation_space.low, b.observation_space.high, b.active_observations):
                    if i == 0 or s not in self.shared_observations or s not in shared_observations:
                        low_limit.append(l)
                        high_limit.append(h)

                    else:
                        pass

                    if s in self.shared_observations and s not in shared_observations:
                        shared_observations.append(s)

                    else:
                        pass

            observation_space = [spaces.Box(low=np.array(low_limit), high=np.array(high_limit), dtype=np.float32)]

        else:
            observation_space = [b.observation_space for b in self.buildings]

        return observation_space

    @property
    def action_space(self) -> List[spaces.Box]:
        """Controller(s) action spaces.

        Returns
        -------
        action_space : List[spaces.Box]
            List of agent(s) action spaces.
        
        Notes
        -----
        If `central_agent` is True, a list of 1 `spaces.Box` object is returned that contains all buildings' limits with the limits in the same order as `buildings`. 
        If `central_agent` is False, a list of `space.Box` objects as many as `buildings` is returned in the same order as `buildings`.
        """

        if self.central_agent:
            low_limit = [v for b in self.buildings for v in b.action_space.low]
            high_limit = [v for b in self.buildings for v in b.action_space.high]
            action_space = [spaces.Box(low=np.array(low_limit), high=np.array(high_limit), dtype=np.float32)]
        else:
            action_space = [b.action_space for b in self.buildings]

        return action_space

    @property
    def observations(self) -> List[List[float]]:
        """Observations at current time step.
        
        Notes
        -----
        If `central_agent` is True, a list of 1 sublist containing all building observation values is returned in the same order as `buildings`. 
        The `shared_observations` values are only included in the first building's observation values. If `central_agent` is False, a list of sublists 
        is returned where each sublist is a list of 1 building's observation values and the sublist in the same order as `buildings`.
        """
        if self._observations_cache is not None and self._observations_cache_time_step == self.time_step:
            return self._observations_cache

        building_observations = [
            b.observations(
                normalize=False,
                periodic_normalization=False,
                check_limits=self.check_observation_limits,
            ) for b in self.buildings
        ]

        if self.central_agent:
            observations = []
            shared_observations = set()
            shared_observation_names = self._shared_observations_set

            for i, b_observations in enumerate(building_observations):
                for k, v in b_observations.items():
                    if i == 0 or k not in shared_observation_names or k not in shared_observations:
                        observations.append(v)

                    if k in shared_observation_names:
                        shared_observations.add(k)

            observations = [observations]

        else:
            observations = [list(o.values()) for o in building_observations]

        self._observations_cache = observations
        self._observations_cache_time_step = self.time_step

        return observations

    @property
    def observation_names(self) -> List[List[str]]:
        """Names of returned observations.

        Notes
        -----
        If `central_agent` is True, a list of 1 sublist containing all building observation names is returned in the same order as `buildings`. 
        The `shared_observations` names are only included in the first building's observation names. If `central_agent` is False, a list of sublists 
        is returned where each sublist is a list of 1 building's observation names and the sublist in the same order as `buildings`.
        """

        if self.central_agent:
            observation_names = []

            for i, b in enumerate(self.buildings):
                for k, _ in b.observations(normalize=False, periodic_normalization=False).items():
                    if i == 0 or k not in self.shared_observations or k not in observation_names:
                        observation_names.append(k)

                    else:
                        pass

            observation_names = [observation_names]

        else:
            observation_names = [list(b.observations().keys()) for b in self.buildings]

        return observation_names

    @property
    def action_names(self) -> List[List[str]]:
        """Names of received actions.

        Notes
        -----
        If `central_agent` is True, a list of 1 sublist containing all building action names is returned in the same order as `buildings`. 
        If `central_agent` is False, a list of sublists is returned where each sublist is a list of 1 building's action names and the sublist 
        in the same order as `buildings`.
        """

        if self.central_agent:
            action_names = []

            for b in self.buildings:
                action_names += b.active_actions

            action_names = [action_names]

        else:
            action_names = [b.active_actions for b in self.buildings]

        return action_names

    def _refresh_action_cache(self):
        self._active_actions_cache = [list(b.active_actions) for b in self.buildings]
        self._expected_central_action_count = sum(len(actions) for actions in self._active_actions_cache)

    @property
    def net_electricity_consumption_emission_without_storage_and_partial_load_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_emission_without_storage_and_partial_load_and_pv` time series, in [kg_co2]."""

        return pd.DataFrame([
            b.net_electricity_consumption_emission_without_storage_and_partial_load_and_pv
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_emission_without_storage_and_pv
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_cost_without_storage_and_partial_load_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_cost_without_storage_and_partial_load_and_pv` time series, in [$]."""

        return pd.DataFrame([
            b.net_electricity_consumption_cost_without_storage_and_partial_load_and_pv
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_cost_without_storage_and_pv
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_without_storage_and_partial_load_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_without_storage_and_partial_load_and_pv` time series, in [kWh]."""

        return pd.DataFrame([
            b.net_electricity_consumption_without_storage_and_partial_load_and_pv
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_without_storage_and_pv
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()


    @property
    def net_electricity_consumption_emission_without_storage_and_partial_load(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_emission_without_storage_and_partial_load` time series, in [kg_co2]."""

        return pd.DataFrame([
            b.net_electricity_consumption_emission_without_storage_and_partial_load
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_emission_without_storage
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_cost_without_storage_and_partial_load(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_cost_without_storage_and_partial_load` time series, in [$]."""

        return pd.DataFrame([
            b.net_electricity_consumption_cost_without_storage_and_partial_load
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_cost_without_storage
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_without_storage_and_partial_load(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_without_storage_and_partial_load` time series, in [kWh]."""

        return pd.DataFrame([
            b.net_electricity_consumption_without_storage_and_partial_load
                if isinstance(b, DynamicsBuilding) else b.net_electricity_consumption_without_storage
                    for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_emission_without_storage_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_emission_without_storage_and_pv` time series, in [kg_co2]."""

        return pd.DataFrame([
            b.net_electricity_consumption_emission_without_storage_and_pv
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_cost_without_storage_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_cost_without_storage_and_pv` time series, in [$]."""

        return pd.DataFrame([
            b.net_electricity_consumption_cost_without_storage_and_pv
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_without_storage_and_pv(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_without_storage_and_pv` time series, in [kWh]."""

        return pd.DataFrame([
            b.net_electricity_consumption_without_storage_and_pv
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()


    @property
    def net_electricity_consumption_emission_without_storage(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_emission_without_storage` time series, in [kg_co2]."""

        return pd.DataFrame([
            b.net_electricity_consumption_emission_without_storage
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_cost_without_storage(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_cost_without_storage` time series, in [$]."""

        return pd.DataFrame([
            b.net_electricity_consumption_cost_without_storage
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_without_storage(self) -> np.ndarray:
        """Summed `Building.net_electricity_consumption_without_storage` time series, in [kWh]."""

        return pd.DataFrame([
            b.net_electricity_consumption_without_storage
                for b in self.buildings
        ]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def net_electricity_consumption_emission(self) -> List[float]:
        """Summed `Building.net_electricity_consumption_emission` time series, in [kg_co2]."""

        return self.__net_electricity_consumption_emission

    @property
    def net_electricity_consumption_cost(self) -> List[float]:
        """Summed `Building.net_electricity_consumption_cost` time series, in [$]."""

        return self.__net_electricity_consumption_cost

    @property
    def net_electricity_consumption(self) -> List[float]:
        """Summed `Building.net_electricity_consumption` time series, in [kWh]."""

        return self.__net_electricity_consumption

    @property
    def cooling_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.cooling_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.cooling_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def heating_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.heating_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.heating_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def dhw_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.dhw_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.dhw_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def cooling_storage_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.cooling_storage_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.cooling_storage_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def heating_storage_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.heating_storage_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.heating_storage_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def dhw_storage_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.dhw_storage_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.dhw_storage_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def electrical_storage_electricity_consumption(self) -> np.ndarray:
        """Summed `Building.electrical_storage_electricity_consumption` time series, in [kWh]."""

        return pd.DataFrame([b.electrical_storage_electricity_consumption for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_cooling_device_to_cooling_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_cooling_device_to_cooling_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_cooling_device_to_cooling_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_heating_device_to_heating_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_heating_device_to_heating_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_heating_device_to_heating_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_dhw_device_to_dhw_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_dhw_device_to_dhw_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_dhw_device_to_dhw_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_to_electrical_storage(self) -> np.ndarray:
        """Summed `Building.energy_to_electrical_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_to_electrical_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_cooling_device(self) -> np.ndarray:
        """Summed `Building.energy_from_cooling_device` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_cooling_device for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_heating_device(self) -> np.ndarray:
        """Summed `Building.energy_from_heating_device` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_heating_device for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_dhw_device(self) -> np.ndarray:
        """Summed `Building.energy_from_dhw_device` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_dhw_device for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_to_non_shiftable_load(self) -> np.ndarray:
        """Summed `Building.energy_to_non_shiftable_load` time series, in [kWh]."""

        return pd.DataFrame([b.energy_to_non_shiftable_load for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_cooling_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_cooling_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_cooling_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()


    @property
    def total_self_consumption(self) -> np.ndarray:
        """Total self-consumption from electrical and thermal storage, in [kWh]."""
        return (
            self.energy_from_electrical_storage +
            self.energy_from_cooling_storage +
            self.energy_from_heating_storage +
            self.energy_from_dhw_storage
        )

    @property
    def energy_from_heating_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_heating_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_heating_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_dhw_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_dhw_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_dhw_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def energy_from_electrical_storage(self) -> np.ndarray:
        """Summed `Building.energy_from_electrical_storage` time series, in [kWh]."""

        return pd.DataFrame([b.energy_from_electrical_storage for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def cooling_demand(self) -> np.ndarray:
        """Summed `Building.cooling_demand`, in [kWh]."""

        return pd.DataFrame([b.cooling_demand for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def heating_demand(self) -> np.ndarray:
        """Summed `Building.heating_demand`, in [kWh]."""

        return pd.DataFrame([b.heating_demand for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def dhw_demand(self) -> np.ndarray:
        """Summed `Building.dhw_demand`, in [kWh]."""

        return pd.DataFrame([b.dhw_demand for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def non_shiftable_load(self) -> np.ndarray:
        """Summed `Building.non_shiftable_load`, in [kWh]."""

        return pd.DataFrame([b.non_shiftable_load for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def solar_generation(self) -> np.ndarray:
        """Summed `Building.solar_generation, in [kWh]`."""

        return pd.DataFrame([b.solar_generation for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()

    @property
    def power_outage(self) -> np.ndarray:
        """Time series of number of buildings experiencing power outage."""

        return pd.DataFrame([b.power_outage_signal for b in self.buildings]).sum(axis = 0, min_count = 1).to_numpy()[:self.time_step + 1]

    @schema.setter
    def schema(self, schema: Union[str, Path, Mapping[str, Any]]):
        dataset = DataSet()

        if isinstance(schema, (str, Path)) and os.path.isfile(schema):
            schema_filepath = Path(schema) if isinstance(schema, str) else schema
            schema = FileHandler.read_json(schema)
            schema['root_directory'] = os.path.split(schema_filepath.absolute())[0] if schema['root_directory'] is None \
                else schema['root_directory']
        
        elif isinstance(schema, str) and schema in dataset.get_dataset_names():
            schema = dataset.get_schema(schema)
            schema['root_directory'] = '' if schema['root_directory'] is None else schema['root_directory']
        
        elif isinstance(schema, dict):
            schema = deepcopy(schema)
            schema['root_directory'] = '' if schema['root_directory'] is None else schema['root_directory']
        
        else:
            raise UnknownSchemaError()
        
        self.__schema = schema

    @render_enabled.setter
    def render_enabled(self, enabled: bool):
        self.__render_enabled = bool(enabled)

    @export_kpis_on_episode_end.setter
    def export_kpis_on_episode_end(self, enabled: bool):
        self.__export_kpis_on_episode_end = bool(enabled)

    @root_directory.setter
    def root_directory(self, root_directory: Union[str, Path]):
        self.__root_directory = root_directory

    @buildings.setter
    def buildings(self, buildings: List[Building]):
        self.__buildings = buildings

    @electric_vehicles.setter
    def electric_vehicles(self, electric_vehicles: List[ElectricVehicle]):
        self.__electric_vehicles = electric_vehicles

    @Environment.episode_tracker.setter
    def episode_tracker(self, episode_tracker: EpisodeTracker):
        Environment.episode_tracker.fset(self, episode_tracker)

        for b in self.buildings:
            b.episode_tracker = self.episode_tracker

    @episode_time_steps.setter
    def episode_time_steps(self, episode_time_steps: Union[int, List[Tuple[int, int]]]):
        self.__episode_time_steps = self.episode_tracker.simulation_time_steps if episode_time_steps is None else episode_time_steps

    @rolling_episode_split.setter
    def rolling_episode_split(self, rolling_episode_split: bool):
        self.__rolling_episode_split = False if rolling_episode_split is None else rolling_episode_split

    @random_episode_split.setter
    def random_episode_split(self, random_episode_split: bool):
        self.__random_episode_split = False if random_episode_split is None else random_episode_split

    @reward_function.setter
    def reward_function(self, reward_function: RewardFunction):
        self.__reward_function = reward_function

    @central_agent.setter
    def central_agent(self, central_agent: bool):
        self.__central_agent = central_agent

    @shared_observations.setter
    def shared_observations(self, shared_observations: List[str]):
        self.__shared_observations = self.get_default_shared_observations() if shared_observations is None else shared_observations
        self._shared_observations_set = set(self.__shared_observations)

    @Environment.random_seed.setter
    def random_seed(self, seed: int):
        Environment.random_seed.fset(self, seed)

        for b in self.buildings:
            b.random_seed = self.random_seed

    @Environment.time_step_ratio.setter
    def time_step_ratio(self, time_step_ratio: int):
        Environment.time_step_ratio.fset(self, time_step_ratio)

        for b in self.buildings:
            b.time_step_ratio = self.time_step_ratio        

    def get_metadata(self) -> Mapping[str, Any]:
        return {
            **super().get_metadata(),
            'reward_function': self.reward_function.__class__.__name__,
            'central_agent': self.central_agent,
            'shared_observations': self.shared_observations,
            'community_market': {
                'enabled': self.community_market_enabled,
                'intra_community_sell_ratio': self.community_market_sell_ratio,
                'grid_export_price': self.community_market_grid_export_price,
                'matching_granularity': 'aggregate_building',
            },
            'buildings': [b.get_metadata() for b in self.buildings],
        }

    @staticmethod
    def get_default_shared_observations() -> List[str]:
        """Names of default common observations across all buildings i.e. observations that have the same value irrespective of the building.
        
        Notes
        -----
        May be used to assigned :attr:`shared_observations` value during `CityLearnEnv` object initialization.
        """

        return [
            'month', 'day_type', 'hour', 'minutes', 'daylight_savings_status',
            'outdoor_dry_bulb_temperature', 'outdoor_dry_bulb_temperature_predicted_1',
            'outdoor_dry_bulb_temperature_predicted_2', 'outdoor_dry_bulb_temperature_predicted_3',
            'outdoor_relative_humidity', 'outdoor_relative_humidity_predicted_1',
            'outdoor_relative_humidity_predicted_2', 'outdoor_relative_humidity_predicted_3',
            'diffuse_solar_irradiance', 'diffuse_solar_irradiance_predicted_1',
            'diffuse_solar_irradiance_predicted_2', 'diffuse_solar_irradiance_predicted_3',
            'direct_solar_irradiance', 'direct_solar_irradiance_predicted_1',
            'direct_solar_irradiance_predicted_2', 'direct_solar_irradiance_predicted_3',
            'carbon_intensity', 'electricity_pricing', 'electricity_pricing_predicted_1',
            'electricity_pricing_predicted_2', 'electricity_pricing_predicted_3',
        ]

    def step(self, actions: List[List[float]]) -> Tuple[List[List[float]], List[float], bool, bool, dict]:
        """Apply actions and advance the environment by one transition."""

        return self._runtime_service.step(actions)

    def get_info(self) -> Mapping[Any, Any]:
        """Other information to return from the `citylearn.CityLearnEnv.step` function."""

        return {}

    def _maybe_log_periodic_metrics(self):
        """Lightweight periodic metrics logging for long training runs."""

        interval = max(0, int(self.metrics_log_interval))

        if interval == 0:
            return

        if self.time_step <= 0:
            return

        if self.time_step % interval != 0 and not self.terminated:
            return

        idx = min(self.time_step - 1, len(self.__net_electricity_consumption) - 1)

        if idx < 0:
            return

        LOGGER.info(
            "Episode %s Step %s/%s | net_kwh=%.5f cost=%.5f co2=%.5f",
            self.episode_tracker.episode,
            self.time_step,
            self.time_steps - 1,
            float(self.__net_electricity_consumption[idx]),
            float(self.__net_electricity_consumption_cost[idx]),
            float(self.__net_electricity_consumption_emission[idx]),
        )

    def _parse_actions(self, actions: List[List[float]]) -> List[Mapping[str, float]]:
        """Compatibility wrapper for runtime action parsing service."""

        return self._runtime_service.parse_actions(actions)

    def evaluate(self, control_condition: EvaluationCondition = None, baseline_condition: EvaluationCondition = None, comfort_band: float = None) -> pd.DataFrame:
        r"""Evaluate cost functions at current time step."""

        return self._kpi_service.evaluate(
            control_condition=control_condition,
            baseline_condition=baseline_condition,
            comfort_band=comfort_band,
            evaluation_condition_cls=EvaluationCondition,
            dynamics_building_cls=DynamicsBuilding,
        )

    def next_time_step(self):
        r"""Advance all buildings to next `time_step`."""

        return self._runtime_service.next_time_step()

    def associate_chargers_to_electric_vehicles(self):
        r"""Associate charger to its corresponding electric_vehicle based on charger simulation state."""

        return self._runtime_service.associate_chargers_to_electric_vehicles()

    def simulate_unconnected_ev_soc(self):
        """Simulate SOC changes for EVs that are not under charger control at t+1."""

        return self._runtime_service.simulate_unconnected_ev_soc()

    def export_final_kpis(self, model: 'Agent' = None, filepath: str = "exported_kpis.csv"):
        """Export episode KPIs to csv."""

        return self._episode_exporter.export_final_kpis(model=model, filepath=filepath)

    def render(self):
        """Render current state of the environment to CSV outputs."""

        return self._episode_exporter.render()

    def _export_episode_render_data(self, final_index: int):
        """Export full episode render rows in one pass for `render_mode='end'`."""

        return self._episode_exporter.export_episode_render_data(final_index)

    def _save_to_csv(self, filename, data):
        """Compatibility wrapper for tests and internal legacy calls."""

        return self._episode_exporter.save_to_csv(filename, data)

    def _flush_render_buffer(self):
        """Write any buffered render rows to disk."""

        return self._episode_exporter.flush_render_buffer()

    def _write_render_rows(self, filename: str, rows: List[Mapping[str, Any]]):
        """Compatibility wrapper for tests and internal legacy calls."""

        return self._episode_exporter.write_render_rows(filename, rows)

    def _parse_render_start_date(self, start_date: Union[str, datetime.date]) -> datetime.date:
        """Return a valid start date for rendering timestamps."""

        return EpisodeExporter.parse_render_start_date(start_date)

    def _ensure_render_output_dir(self, *, ensure_exists: bool = True):
        """Prepare the render output directory and optionally create it on disk."""

        return self._episode_exporter.ensure_output_dir(ensure_exists=ensure_exists)

    def _get_iso_timestamp(self):
        return self._episode_exporter.get_iso_timestamp()

    def _override_render_time_step(self, index: int):
        return self._episode_exporter.override_render_time_step(index)

    @staticmethod
    def _restore_render_time_step(snapshot):
        return EpisodeExporter.restore_render_time_step(snapshot)

    def _reset_time_tracking(self):
        return self._episode_exporter.reset_time_tracking()

    def reset(self, seed: int = None, options: Mapping[str, Any] = None) -> Tuple[List[List[float]], dict]:
        r"""Reset `CityLearnEnv` to initial state.

        Parameters
        ----------
        seed: int, optional
            Use to updated :code:`citylearn.CityLearnEnv.random_seed` if value is provided.
        options: Mapping[str, Any], optional
            Use to pass additional data to environment on reset. Not used in this base class
            but included to conform to gymnasium interface.
        
        Returns
        -------
        observations: List[List[float]]
            :attr:`observations`.
        info: dict
            A dictionary that may contain additional information regarding the reason for a `terminated` signal.
            `info` contains auxiliary diagnostic information (helpful for debugging, learning, and logging).
            Override :meth"`get_info` to get custom key-value pairs in `info`.
        """

        # object reset
        super().reset()
        self._final_kpis_exported = False

        # update seed
        if seed is not None:
            self.random_seed = seed
        else:
            pass

        # update time steps for time series
        self.episode_tracker.next_episode(
            self.episode_time_steps,
            self.rolling_episode_split,
            self.random_episode_split,
            self.random_seed,
        )

        for building in self.buildings:
            building.reset()

        for ev in self.electric_vehicles:
            ev.reset()

        self.associate_chargers_to_electric_vehicles()

        # reset reward function (does nothing by default)
        self.reward_function.reset()

        # variable reset
        self.__rewards = [[]]
        self.__net_electricity_consumption = []
        self.__net_electricity_consumption_cost = []
        self.__net_electricity_consumption_emission = []
        self._last_community_market_settlement = []
        self._community_market_settlement_history = []
        self._observations_cache = None
        self._observations_cache_time_step = -1
        episode_index = int(getattr(self.episode_tracker, 'episode', 0))
        self._ev_drift_random_state = np.random.RandomState(int(self.random_seed) + episode_index)
        self._render_buffer.clear()
        self._refresh_action_cache()
        self.update_variables()

        return self.observations, self.get_info()

    def _configure_community_market(self):
        config = {}
        if isinstance(self.schema, dict):
            config = self.schema.get('community_market', {}) or {}

        self.community_market_enabled = parse_bool(
            config.get('enabled', False),
            default=False,
            path='community_market.enabled',
        )
        ratio = config.get('intra_community_sell_ratio', 0.8)

        try:
            ratio = float(ratio)
        except (TypeError, ValueError):
            ratio = 0.8

        self.community_market_sell_ratio = min(max(ratio, 0.0), 1.0)
        self.community_market_grid_export_price = config.get('grid_export_price', 0.0)

    def update_variables(self):
        """Update district-level aggregate variables."""

        return self._runtime_service.update_variables()

    def load_agent(self, agent: Union[str, 'Agent'] = None, **kwargs) -> Union[Any, 'Agent']:
        """Return :class:`Agent` or sub class object as defined by the `schema`.

        Parameters
        ----------
        agent: Union[str, 'citylearn.agents.base.Agent], optional
            Agent class or string describing path to agent class, e.g. 'citylearn.agents.base.BaselineAgent'.
            If a value is not provided, defaults to the agent defined in the schema:agent:type.

        **kwargs : dict
            Agent initialization attributes. For most agents e.g. CityLearn and Stable-Baselines3 agents, 
            an intialized :py:attr:`env` must be parsed to the agent :py:meth:`init` function.
        
        Returns
        -------
        agent: Agent
            Initialized agent.
        """

        # set agent class
        if agent is not None:
            agent_type = agent

            if not isinstance(agent_type, str):
                agent_type = [agent_type.__module__] + [agent_type.__name__]
                agent_type = '.'.join(agent_type)

            else:
                pass

        # set agent init attributes
        else:
            agent_type = self.schema['agent']['type']

        if kwargs is not None and len(kwargs) > 0:
            agent_attributes = dict(kwargs)

        elif agent is None:
            agent_attributes = dict(self.schema['agent'].get('attributes', {}))

        else:
            agent_attributes = {}

        if 'env' not in agent_attributes:
            agent_attributes['env'] = self

        agent_module = '.'.join(agent_type.split('.')[0:-1])
        agent_name = agent_type.split('.')[-1]
        agent_constructor = getattr(importlib.import_module(agent_module), agent_name)
        agent = agent_constructor(**agent_attributes)

        return agent

    def _load(self, schema: Mapping[str, Any], **kwargs) -> Tuple[Union[Path, str], List[Building], List[ElectricVehicle], Union[int, List[Tuple[int, int]]], bool, bool, float, RewardFunction, bool, List[str], EpisodeTracker]:
        """Compatibility wrapper for schema loading service."""

        return self._loading_service.load(schema, **kwargs)

    def _load_building(self, index: int, building_name: str, schema: dict, episode_tracker: EpisodeTracker, pv_sizing_data: pd.DataFrame, battery_sizing_data: pd.DataFrame, **kwargs) -> Building:
        """Compatibility wrapper for building loading service."""

        return self._loading_service.load_building(
            index,
            building_name,
            schema,
            episode_tracker,
            pv_sizing_data,
            battery_sizing_data,
            **kwargs,
        )

    def process_metadata(self, schema, building_schema, chargers_list, washing_machines_list, index, energy_simulation: EnergySimulation, **kwargs):
        """Compatibility wrapper for metadata processing service."""

        return self._loading_service.process_metadata(
            schema,
            building_schema,
            chargers_list,
            washing_machines_list,
            index,
            energy_simulation,
            **kwargs,
        )

    def _load_electric_vehicle(self, electric_vehicle_name: str, schema: dict, electric_vehicle_schema: dict, episode_tracker: EpisodeTracker, time_step_ratio) -> ElectricVehicle:
        """Compatibility wrapper for electric vehicle loading service."""

        return self._loading_service.load_electric_vehicle(
            electric_vehicle_name,
            schema,
            electric_vehicle_schema,
            episode_tracker,
            time_step_ratio,
        )

    def _load_washing_machine(
        self,
        washing_machine_name: str,
        schema: dict,
        washing_machine_schema: dict,
        episode_tracker: EpisodeTracker
    ) -> WashingMachine:
        """Compatibility wrapper for washing machine loading service."""

        return self._loading_service.load_washing_machine(
            washing_machine_name,
            schema,
            washing_machine_schema,
            episode_tracker,
        )

    def __str__(self) -> str:
        """
        Return a string representation of the current simulation state.

        Useful for logging or quick inspection of internal values.
        """
        return str(self.as_dict())

    def as_dict(self) -> dict:
        """
        Convert the current simulation state to a dictionary.

        This includes key performance indicators such as energy usage, emissions, 
        and electricity pricing at the current time step.

        Returns
        -------
        dict
            Dictionary with energy and environmental metrics for the current step.
        """
        if len(self.net_electricity_consumption) == 0:
            idx = 0
        else:
            idx = max(0, min(self.time_step, len(self.net_electricity_consumption) - 1))

        def _safe_value(series, index: int) -> float:
            return float(series[index]) if 0 <= index < len(series) else 0.0

        self_consumption = 0.0
        stored_energy = 0.0
        total_solar_generation = 0.0

        for building in self.buildings:
            self_consumption += (
                _safe_value(building.energy_from_electrical_storage, idx)
                + _safe_value(building.energy_from_cooling_storage, idx)
                + _safe_value(building.energy_from_heating_storage, idx)
                + _safe_value(building.energy_from_dhw_storage, idx)
            )
            stored_energy += _safe_value(building.energy_to_electrical_storage, idx)
            total_solar_generation += _safe_value(building.solar_generation, idx)

        return {
            "Net Electricity Consumption-kWh": _safe_value(self.net_electricity_consumption, idx),
            "Self Consumption-kWh": self_consumption,
            "Stored energy by community- kWh": stored_energy,
            "Total Solar Generation-kWh": total_solar_generation,
            "CO2-kg_co2": _safe_value(self.net_electricity_consumption_emission, idx),
            "Price-$": _safe_value(self.net_electricity_consumption_cost, idx),
        }
class Error(Exception):
    """Base class for other exceptions."""

class UnknownSchemaError(Error):
    """Raised when a schema is not a data set name, dict nor filepath."""
    __MESSAGE = 'Unknown schema parsed into constructor. Schema must be name of CityLearn data set,'\
        ' a filepath to JSON representation or `dict` object of a CityLearn schema.'\
        ' Call citylearn.data.DataSet.get_names() for list of available CityLearn data sets.'

    def __init__(self,message=None):
        super().__init__(self.__MESSAGE if message is None else message)



import pickle
from typing import Any, Iterable, Union
import numpy as np
import simplejson as json
import yaml


def parse_bool(value: Any, default: Any = None, path: str = 'value'):
    """Parse boolean-like values from schema/config sources.

    Accepted inputs:
    - bool
    - int/np.integer: 0 or 1
    - float/np.floating: 0.0 or 1.0
    - str (case-insensitive): true/false, 1/0, yes/no, on/off

    Parameters
    ----------
    value: Any
        Value to parse.
    default: Any, optional
        Fallback value if `value` is None.
    path: str, default: 'value'
        Config path used in validation error messages.
    """

    if value is None:
        if default is None:
            return None
        return parse_bool(default, default=None, path=path)

    if isinstance(value, (bool, np.bool_)):
        return bool(value)

    if isinstance(value, (int, np.integer)) and not isinstance(value, bool):
        if int(value) in (0, 1):
            return bool(int(value))
        raise ValueError(f"{path} must be one of: true/false, 1/0, yes/no, on/off.")

    if isinstance(value, (float, np.floating)):
        if float(value) in (0.0, 1.0):
            return bool(int(float(value)))
        raise ValueError(f"{path} must be one of: true/false, 1/0, yes/no, on/off.")

    if isinstance(value, str):
        token = value.strip().lower()
        if token in {'true', '1', 'yes', 'on'}:
            return True
        if token in {'false', '0', 'no', 'off'}:
            return False
        raise ValueError(f"{path} must be one of: true/false, 1/0, yes/no, on/off.")

    raise ValueError(f"{path} must be one of: true/false, 1/0, yes/no, on/off.")


class FileHandler:
    @staticmethod
    def read_json(filepath: str, **kwargs) -> dict:
        """Return JSON document as dictionary.
        
        Parameters
        ----------
        filepath : str
        pathname of JSON document.

        Other Parameters
        ----------------
        **kwargs : dict
            Other infrequently used keyword arguments to be parsed to `simplejson.load`.
        
        Returns
        -------
        dict
            JSON document converted to dictionary.
        """

        with open(filepath) as f:
            json_file = json.load(f, **kwargs)

        return json_file

    @staticmethod
    def write_json(filepath: str, dictionary: dict, **kwargs):
        """Write dictionary to JSON file.
        
        Parameters
        ----------
        filepath : str
            pathname of JSON document.
        dictionary: dict
            dictionary to convert to JSON.

        Other Parameters
        ----------------
        **kwargs : dict
            Other infrequently used keyword arguments to be parsed to `simplejson.dump`.
        """

        kwargs = {'ignore_nan': True, 'sort_keys': False, 'default': str, 'indent': 2, **kwargs}
        
        with open(filepath,'w') as f:
            json.dump(dictionary, f, **kwargs)

    @staticmethod
    def read_yaml(filepath: str) -> dict:
        """Return YAML document as dictionary.
        
        Parameters
        ----------
        filepath : str
        pathname of YAML document.
        
        Returns
        -------
        dict
            YAML document converted to dictionary.
        """

        with open(filepath, 'r') as f:
            data = yaml.safe_load(f)

        return data

    @staticmethod
    def write_yaml(filepath: str, dictionary: dict, **kwargs):
        """Write dictionary to YAML file. 
        
        Parameters
        ----------
        filepath : str
            pathname of YAML document.
        dictionary: dict
            dictionary to convert to YAML.

        Other Parameters
        ----------------
        **kwargs : dict
            Other infrequently used keyword arguments to be parsed to `yaml.safe_dump`.
        """

        kwargs = {'sort_keys': False, 'indent': 2, **kwargs}
        
        with open(filepath, 'w') as f:
            yaml.safe_dump(dictionary, f, **kwargs)

    @staticmethod
    def read_pickle(filepath: str, **kwargs) -> Any:
        """Return pickle file as some Python class object.
        
        Parameters
        ----------
        filepath : str
        pathname of pickle file.

        Other Parameters
        ----------------
        **kwargs : dict
            Other infrequently used keyword arguments to be parsed to `pickle.load`.
        
        Returns
        -------
        Any
            Pickle file as a Python object.
        """

        with (open(filepath, 'rb')) as f:
            data = pickle.load(f, **kwargs)

        return data

    @staticmethod
    def write_pickle(filepath: str, data: Any, **kwargs):
        """Write Python object to pickle file. 
        
        Parameters
        ----------
        filepath : str
            pathname of pickle document.
        data: dict
            object to convert to pickle.

        Other Parameters
        ----------------
        **kwargs : dict
            Other infrequently used keyword arguments to be parsed to `pickle.dump`.
        """

        with open(filepath, 'wb') as f:
            pickle.dump(data, f, **kwargs)

    @staticmethod
    def join_url(*args: str) -> str:
        url = '/'.join([a.strip('/') for a in args])

        return url
    
class NoiseUtils:
    @staticmethod
    def generate_gaussian_noise(input_data: Union[np.ndarray, Iterable[float]], noise_std: float) -> np.ndarray:
        """Generates Gaussian noise matching input shape.
        
        Parameters
        ----------
        input_data : Union[np.ndarray, Iterable[float]]
            Time series to add noise to.
        noise_std : float
            Noise standard deviation (ignored if <= 0)
            
        Returns
        -------
            noise: np.ndarray
                Zero-mean noise array with same shape as input
        """

        arr = np.asarray(input_data)  # Handles both ndarray and Iterable
        if noise_std <= 0:
            return np.zeros(arr.shape)
        return np.random.normal(loc=0, scale=noise_std, size=arr.shape)

    @staticmethod
    def generate_scaled_noise(input_data: Union[np.ndarray, Iterable[float]], noise_std: float, scale: float = 1.0) -> np.ndarray:
        """Generates pre-scaled noise (e.g., for percentage values)."""
        
        return NoiseUtils.generate_gaussian_noise(input_data, noise_std) * scale



from __future__ import annotations

from collections import defaultdict
import csv
import datetime
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Mapping, TYPE_CHECKING, Union

import numpy as np

if TYPE_CHECKING:
    from citylearn.agents.base import Agent
    from citylearn.citylearn import CityLearnEnv
    from citylearn.electric_vehicle import ElectricVehicle

LOGGER = logging.getLogger(__name__)


class EpisodeExporter:
    """Internal helper that owns rendering/export behaviour for ``CityLearnEnv``."""

    DEFAULT_RENDER_START_DATE = datetime.date(2024, 1, 1)

    def __init__(self, env: "CityLearnEnv"):
        self.env = env

    @staticmethod
    def parse_render_start_date(start_date: Union[str, datetime.date, datetime.datetime]) -> datetime.date:
        """Return a valid start date for rendering timestamps."""

        if start_date is None:
            return EpisodeExporter.DEFAULT_RENDER_START_DATE

        if isinstance(start_date, datetime.datetime):
            return start_date.date()

        if isinstance(start_date, datetime.date):
            return start_date

        if isinstance(start_date, str):
            try:
                return datetime.date.fromisoformat(start_date)
            except ValueError as exc:
                raise ValueError(
                    "CityLearnEnv start_date must be in ISO format 'YYYY-MM-DD'."
                ) from exc

        raise TypeError(
            "CityLearnEnv start_date must be a date, datetime, or ISO format string."
        )

    def export_final_kpis(self, model: "Agent" = None, filepath: str = "exported_kpis.csv"):
        """Export episode KPIs to csv."""

        env = self.env
        self.ensure_output_dir()
        file_path = os.path.join(env.new_folder_path, filepath)

        if model is not None and getattr(model, 'env', None) is not None:
            kpis = model.env.evaluate()
        else:
            kpis = env.evaluate()

        kpis = kpis.pivot(index='cost_function', columns='name', values='value').round(3)
        kpis = kpis.fillna('')
        kpis = kpis.reset_index()
        kpis = kpis.rename(columns={'cost_function': 'KPI'})
        kpis.to_csv(file_path, index=False, encoding='utf-8')
        env._final_kpis_exported = True

    def render(self):
        """Render one time step to CSV outputs."""

        env = self.env

        if not getattr(env, 'render_enabled', False):
            return

        if env.render_mode == 'end' and getattr(env, '_defer_render_flush', False):
            return

        if env.render_mode == 'end' and (env.terminated or env.truncated):
            return

        self.ensure_output_dir()
        iso_timestamp = self.get_iso_timestamp()
        os.makedirs(env.new_folder_path, exist_ok=True)

        episode_num = env.episode_tracker.episode

        self.save_to_csv(
            f"exported_data_community_ep{episode_num}.csv",
            {"timestamp": iso_timestamp, **env.as_dict()},
        )

        for building in env.buildings:
            self.save_to_csv(
                f"exported_data_{building.name.lower()}_ep{episode_num}.csv",
                {"timestamp": iso_timestamp, **building.as_dict()},
            )

            battery = building.electrical_storage
            self.save_to_csv(
                f"exported_data_{building.name.lower()}_battery_ep{episode_num}.csv",
                {"timestamp": iso_timestamp, **battery.as_dict()},
            )

            for charger in building.electric_vehicle_chargers or []:
                self.save_to_csv(
                    f"exported_data_{building.name.lower()}_{charger.charger_id}_ep{episode_num}.csv",
                    {"timestamp": iso_timestamp, **charger.as_dict()},
                )

        self.save_to_csv(
            f"exported_data_pricing_ep{episode_num}.csv",
            {"timestamp": iso_timestamp, **env.buildings[0].pricing.as_dict(env.time_step)},
        )

        for ev in env.electric_vehicles:
            self.save_to_csv(
                f"exported_data_{ev.name.lower()}_ep{episode_num}.csv",
                {"timestamp": iso_timestamp, **ev.as_dict()},
            )

    def _set_charger_render_state(self, charger, time_step: int, ev_lookup: Mapping[str, "ElectricVehicle"]):
        """Set charger connected/incoming EV pointers to match schedule at a given time step."""

        sim = charger.charger_simulation
        state = sim.electric_vehicle_charger_state[time_step] if time_step < len(sim.electric_vehicle_charger_state) else np.nan
        ev_id = sim.electric_vehicle_id[time_step] if time_step < len(sim.electric_vehicle_id) else None
        valid_ev_id = isinstance(ev_id, str) and ev_id.strip() not in {"", "nan"}

        connected_ev = ev_lookup.get(ev_id) if valid_ev_id and state == 1 else None
        incoming_ev = ev_lookup.get(ev_id) if valid_ev_id and state == 2 else None

        charger.connected_electric_vehicle = connected_ev
        charger.incoming_electric_vehicle = incoming_ev

    def export_episode_render_data(self, final_index: int):
        """Export full episode render rows in one pass for ``render_mode='end'``."""

        env = self.env

        if final_index < 0:
            return

        self.ensure_output_dir()
        episode_num = env.episode_tracker.episode
        rows_by_filename: Dict[str, List[Mapping[str, Any]]] = defaultdict(list)
        ev_lookup = {ev.name: ev for ev in env.electric_vehicles}
        original_charger_state = {}
        time_step_snapshot = self.override_render_time_step(0)
        original_year = env.year
        original_day = env.current_day
        original_start_datetime = getattr(env, '_render_start_datetime', None)

        try:
            self.reset_time_tracking()

            for t in range(final_index + 1):
                for obj, _ in time_step_snapshot:
                    try:
                        obj.time_step = t
                    except AttributeError:
                        pass

                timestamp = self.get_iso_timestamp()
                rows_by_filename[f"exported_data_community_ep{episode_num}.csv"].append(
                    {"timestamp": timestamp, **env.as_dict()}
                )

                for building in env.buildings:
                    rows_by_filename[f"exported_data_{building.name.lower()}_ep{episode_num}.csv"].append(
                        {"timestamp": timestamp, **building.as_dict()}
                    )
                    battery = building.electrical_storage
                    rows_by_filename[f"exported_data_{building.name.lower()}_battery_ep{episode_num}.csv"].append(
                        {"timestamp": timestamp, **battery.as_dict()}
                    )

                    for charger in building.electric_vehicle_chargers or []:
                        if charger not in original_charger_state:
                            original_charger_state[charger] = (
                                charger.connected_electric_vehicle,
                                charger.incoming_electric_vehicle,
                            )
                        self._set_charger_render_state(charger, t, ev_lookup)
                        rows_by_filename[f"exported_data_{building.name.lower()}_{charger.charger_id}_ep{episode_num}.csv"].append(
                            {"timestamp": timestamp, **charger.as_dict()}
                        )

                rows_by_filename[f"exported_data_pricing_ep{episode_num}.csv"].append(
                    {"timestamp": timestamp, **env.buildings[0].pricing.as_dict(t)}
                )

                for ev in env.electric_vehicles:
                    rows_by_filename[f"exported_data_{ev.name.lower()}_ep{episode_num}.csv"].append(
                        {"timestamp": timestamp, **ev.as_dict()}
                    )

        finally:
            for charger, state in original_charger_state.items():
                charger.connected_electric_vehicle, charger.incoming_electric_vehicle = state

            self.restore_render_time_step(time_step_snapshot)
            env.year = original_year
            env.current_day = original_day
            env._render_start_datetime = original_start_datetime

        for filename, rows in rows_by_filename.items():
            file_path = Path(env.new_folder_path) / filename
            if file_path.exists():
                file_path.unlink()
            self.write_render_rows(filename, rows)

    def save_to_csv(self, filename: str, data: Mapping[str, Any]):
        """Save one render row to CSV."""

        env = self.env

        if env._buffer_render and getattr(env, '_defer_render_flush', False):
            env._render_buffer[filename].append(dict(data))
            return

        self.write_render_rows(filename, [dict(data)])

    def flush_render_buffer(self):
        """Write any buffered render rows to disk."""

        env = self.env

        if not getattr(env, '_render_buffer', None):
            return

        has_pending_rows = any(env._render_buffer.values())
        if not has_pending_rows:
            env._render_buffer.clear()
            return

        try:
            target_dir = Path(env.new_folder_path)
        except Exception:
            target_dir = None

        if target_dir is not None:
            LOGGER.info("Writing buffered render exports to %s ...", target_dir)

        original_defer = env._defer_render_flush
        original_buffer_state = env._buffer_render
        env._defer_render_flush = False
        env._buffer_render = False

        try:
            for filename, rows in list(env._render_buffer.items()):
                if rows:
                    self.write_render_rows(filename, rows)
        finally:
            env._render_buffer.clear()
            env._buffer_render = original_buffer_state
            env._defer_render_flush = original_defer

    def write_render_rows(self, filename: str, rows: List[Mapping[str, Any]]):
        """Write one or more render rows to disk with minimal rewrites."""

        env = self.env
        file_path = Path(env.new_folder_path) / filename
        file_path.parent.mkdir(parents=True, exist_ok=True)

        if not rows:
            return

        buffered_fieldnames = list(dict.fromkeys(field for row in rows for field in row.keys()))

        if not file_path.exists():
            fieldnames = buffered_fieldnames
            with file_path.open('w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                writer.writeheader()
                for row in rows:
                    writer.writerow({field: row.get(field, '') for field in fieldnames})
            return

        with file_path.open('r', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            existing_rows = list(reader)
            existing_fieldnames = reader.fieldnames or []

        missing_fieldnames = [field for field in buffered_fieldnames if field not in existing_fieldnames]
        if missing_fieldnames:
            extended_fieldnames = [*existing_fieldnames, *missing_fieldnames]
            with file_path.open('w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=extended_fieldnames)
                writer.writeheader()
                for row in existing_rows:
                    writer.writerow({field: row.get(field, '') for field in extended_fieldnames})
                for row in rows:
                    writer.writerow({field: row.get(field, '') for field in extended_fieldnames})
            return

        with file_path.open('a', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=existing_fieldnames)
            for row in rows:
                writer.writerow({field: row.get(field, '') for field in existing_fieldnames})

    def ensure_output_dir(self, *, ensure_exists: bool = True):
        """Prepare the render output directory and optionally create it on disk."""

        env = self.env
        base_render_path = Path(
            getattr(env, 'render_output_root', Path(__file__).resolve().parents[1] / 'render_logs')
        ).expanduser()

        if ensure_exists:
            try:
                base_render_path.mkdir(parents=True, exist_ok=True)
            except PermissionError:
                fallback = (Path.cwd() / 'render_logs').resolve()
                fallback.mkdir(parents=True, exist_ok=True)
                env.render_output_root = fallback
                base_render_path = fallback

        render_dir = getattr(env, '_render_directory_path', None)
        needs_new_dir = render_dir is None

        if not needs_new_dir and ensure_exists:
            render_dir = Path(render_dir)
            try:
                needs_new_dir = not render_dir.is_relative_to(base_render_path)
            except AttributeError:
                needs_new_dir = base_render_path not in render_dir.parents and render_dir != base_render_path

        if needs_new_dir:
            if env.render_session_name:
                render_dir = (base_render_path / Path(env.render_session_name)).expanduser().resolve()
            else:
                if getattr(env, '_render_timestamp', None) is None:
                    env._render_timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                render_dir = (base_render_path / env._render_timestamp).resolve()

            env._render_directory_path = render_dir
        else:
            render_dir = Path(env._render_directory_path)

        if ensure_exists:
            render_dir.mkdir(parents=True, exist_ok=True)
            if not env._render_dir_initialized:
                if env.render_session_name:
                    for csv_file in render_dir.glob('exported_*.csv'):
                        try:
                            csv_file.unlink()
                        except OSError:
                            pass
                env._render_dir_initialized = True

        env.new_folder_path = str(render_dir)

    def get_iso_timestamp(self) -> str:
        """Return current episode timestamp string in ISO format."""

        env = self.env

        if env.time_step == 0 or getattr(env, '_render_start_datetime', None) is None:
            self.reset_time_tracking()

        start_datetime = env._render_start_datetime
        timestamp_dt = start_datetime + datetime.timedelta(seconds=env.time_step * env.seconds_per_time_step)
        env.year = timestamp_dt.year
        env.current_day = timestamp_dt.day

        return timestamp_dt.strftime("%Y-%m-%dT%H:%M:%S")

    def override_render_time_step(self, index: int):
        """Temporarily set time_step to `index` for the environment and descendants."""

        env = self.env
        snapshot = []

        def _record(obj):
            if hasattr(obj, 'time_step'):
                snapshot.append((obj, obj.time_step))
                obj.time_step = index

        _record(env)
        for building in getattr(env, 'buildings', []):
            _record(building)
            electrical_storage = getattr(building, 'electrical_storage', None)
            if electrical_storage is not None:
                _record(electrical_storage)

            for charger in getattr(building, 'electric_vehicle_chargers', []) or []:
                _record(charger)

            for washing_machine in getattr(building, 'washing_machines', []) or []:
                _record(washing_machine)

        for ev in getattr(env, 'electric_vehicles', []):
            _record(ev)
            battery = getattr(ev, 'battery', None)
            if battery is not None:
                _record(battery)

        return snapshot

    @staticmethod
    def restore_render_time_step(snapshot):
        for obj, value in snapshot:
            try:
                obj.time_step = value
            except AttributeError:
                pass

    def reset_time_tracking(self):
        """Reset render timestamp tracking to episode start."""

        env = self.env
        start_offset = getattr(env.episode_tracker, 'episode_start_time_step', 0)
        base_datetime = datetime.datetime.combine(env.render_start_date, datetime.time())
        base_datetime += datetime.timedelta(seconds=start_offset * env.seconds_per_time_step)
        env._render_start_datetime = base_datetime
        env.year = base_datetime.year
        env.current_day = base_datetime.day



from copy import deepcopy
from typing import Any, List, Tuple
import numpy as np
from citylearn.agents.rbc import RBC
from citylearn.agents.sac import SACRBC
from citylearn.citylearn import CityLearnEnv

try:
    from sklearn.decomposition import PCA
    from sklearn.linear_model import LinearRegression
except (ModuleNotFoundError, ImportError) as e:
    raise Exception("This functionality requires you to install scikit-learn. You can install scikit-learn by : pip scikit-learn, or for more detailed instructions please visit https://scikit-learn.org/stable/install.html.")

try:
    import torch
except (ModuleNotFoundError, ImportError) as e:
    raise Exception("This functionality requires you to install torch. You can install torch by : pip install torch torchvision, or for more detailed instructions please visit https://pytorch.org.")

from citylearn.agents.sac import SAC
from citylearn.preprocessing import Encoder, NoNormalization, PeriodicNormalization, RemoveFeature
from citylearn.rl import RegressionBuffer

class MARLISA(SAC):
    __COORDINATION_VARIABLE_COUNT = 2

    def __init__(
        self, *args, regression_buffer_capacity: int = None, start_regression_time_step: int = None, 
        regression_frequency: int = None, information_sharing: bool = None, pca_compression: float = None, iterations: int = None, **kwargs
    ):
        self.__coordination_variables_history = None
        self.information_sharing = information_sharing

        super().__init__(*args, **kwargs)

        self.regression_buffer_capacity = regression_buffer_capacity
        self.start_regression_time_step = start_regression_time_step
        self.pca_compression = pca_compression
        self.iterations = iterations
        self.regression_frequency = regression_frequency

        # internally defined
        self.regression_buffer = [RegressionBuffer(int(self.regression_buffer_capacity)) for _ in self.action_space]
        self.state_estimator = [LinearRegression() for _ in self.action_space]
        self.pca = [None for _ in self.action_space]
        self.pca_flag = [False for _ in self.action_dimension]
        self.regression_flag = [0 for _ in self.action_dimension]
        self.energy_size_coefficient = None
        self.total_coefficient = None
        self.regression_encoders = self.set_regression_encoders()
        self.set_energy_coefficients()
        self.set_pca()

    @property
    def regression_buffer_capacity(self) -> int:
        return self.__regression_buffer_capacity

    @property
    def start_regression_time_step(self) -> int:
        return self.__start_regression_time_step

    @property
    def regression_frequency(self) -> int:
        return self.__regression_frequency

    @property
    def information_sharing(self) -> bool:
        return self.__information_sharing

    @property
    def pca_compression(self) -> float:
        return self.__pca_compression

    @property
    def iterations(self) -> int:
        return self.__iterations

    @property
    def coordination_variables_history(self) -> List[float]:
        return self.__coordination_variables_history

    @SAC.hidden_dimension.setter
    def hidden_dimension(self, hidden_dimension: List[float]):
        hidden_dimension = [400, 300] if hidden_dimension is None else hidden_dimension
        SAC.hidden_dimension.fset(self, hidden_dimension)

    @SAC.batch_size.setter
    def batch_size(self, batch_size: int):
        batch_size = 100 if batch_size is None else batch_size
        SAC.batch_size.fset(self, batch_size)

    @regression_buffer_capacity.setter
    def regression_buffer_capacity(self, regression_buffer_capacity: int):
        self.__regression_buffer_capacity = 3e4 if regression_buffer_capacity is None else regression_buffer_capacity

    @regression_frequency.setter
    def regression_frequency(self, regression_frequency: int):
        self.__regression_frequency = 2500 if regression_frequency is None else regression_frequency

    @start_regression_time_step.setter
    def start_regression_time_step(self, start_regression_time_step: int):
        default_time_step = 2
        start_regression_time_step = default_time_step if start_regression_time_step is None else start_regression_time_step
        assert start_regression_time_step < self.standardize_start_time_step, 'start_regression_time_step must be < standardize_start_time_step'
        self.__start_regression_time_step = start_regression_time_step

    @information_sharing.setter
    def information_sharing(self, information_sharing: bool):
        self.__information_sharing = True if information_sharing is None else information_sharing

    @pca_compression.setter
    def pca_compression(self, pca_compression: float):
        self.__pca_compression = 1.0 if pca_compression is None else pca_compression

    @iterations.setter
    def iterations(self, iterations: int):
        self.__iterations = 2 if iterations is None else iterations

    def update(self, observations: List[List[float]], actions: List[List[float]], reward: List[float], next_observations: List[List[float]], terminated: bool, truncated: bool):
        r"""Update replay buffer.

        Parameters
        ----------
        observations : List[List[float]]
            Previous time step observations.
        actions : List[List[float]]
            Previous time step actions.
        reward : List[float]
            Current time step reward.
        next_observations : List[List[float]]
            Current time step observations.
        terminated : bool
            Indication that episode has ended.
        truncated : bool
            If episode truncates due to a time limit or a reason that is not defined as part of the task MDP.
        """

        # Run once the regression model has been fitted
        # Normalize all the observations using periodical normalization, one-hot encoding, or -1, 1 scaling. It also removes observations that are not necessary (solar irradiance if there are no solar PV panels).

        for i, (o, a, r, n, c0, c1) in enumerate(zip(observations, actions, reward, next_observations, self.coordination_variables_history[0], self.coordination_variables_history[1])):
            if self.information_sharing:
                # update regression buffer
                variables = np.hstack(np.concatenate((self.get_encoded_regression_variables(i, o), a)))
                # The targets are the net electricity consumption.
                target = self.get_encoded_regression_targets(i, n)
                self.regression_buffer[i].push(variables, target)
            
            else:
                pass

            if self.regression_flag[i] > 1:
                o = self.get_encoded_observations(i, o)
                n = self.get_encoded_observations(i, n)

                # Only executed during the random exploration phase. Pushes unnormalized tuples into the replay buffer.
                if self.information_sharing:
                    o = np.hstack(np.concatenate((o, c0), dtype=float))
                    n = np.hstack(np.concatenate((n, c1), dtype=float))
                
                else:
                    pass

                # Executed during the training phase. States and rewards pushed into the replay buffer are normalized and processed using PCA.
                if self.pca_flag[i]:
                    o = self.get_normalized_observations(i, o)
                    o = self.pca[i].transform(o.reshape(1, -1))[0]
                    n = self.get_normalized_observations(i, n)
                    n = self.pca[i].transform(n.reshape(1, -1))[0]
                    r = self.get_normalized_reward(i, r)
                else:
                    pass

                self.replay_buffer[i].push(o, a, r, n, terminated)

            else:
                pass

            if self.time_step >= self.start_regression_time_step\
                and (self.regression_flag[i] < 2 or self.time_step%self.regression_frequency == 0):
                if self.information_sharing:
                    self.state_estimator[i].fit(self.regression_buffer[i].x, self.regression_buffer[i].y)
                
                else:
                    pass

                if self.regression_flag[i] < 2:
                    self.regression_flag[i] += 1
                
                else:
                    pass

            else:
                pass

            if self.time_step >= self.standardize_start_time_step and self.batch_size <= len(self.replay_buffer[i]):
                # This code only runs once. Once the random exploration phase is over, we normalize all the states and rewards to make them have mean=0 and std=1, and apply PCA. We push the normalized compressed values back into the buffer, replacing the old buffer.
                if not self.pca_flag[i]:
                    # calculate normalized observations and rewards
                    X = np.array([j[0] for j in self.replay_buffer[i].buffer], dtype=float)
                    self.norm_mean[i] = np.nanmean(X, axis=0)
                    self.norm_std[i] = np.nanstd(X, axis=0) + 1e-5
                    X = self.get_normalized_observations(i, X)
                    self.pca[i].fit(X)

                    R = np.array([j[2] for j in self.replay_buffer[i].buffer], dtype=float)
                    self.r_norm_mean[i] = np.nanmean(R, dtype=float)
                    self.r_norm_std[i] = np.nanstd(R, dtype=float)/self.reward_scaling + 1e-5

                    # update buffer with normalization
                    self.replay_buffer[i].buffer = [(
                        np.hstack(self.pca[i].transform(self.get_normalized_observations(i, o).reshape(1,-1))[0]),
                        a,
                        self.get_normalized_reward(i, r),
                        np.hstack(self.pca[i].transform(self.get_normalized_observations(i, n).reshape(1,-1))[0]),
                        d
                    ) for o, a, r, n, d in self.replay_buffer[i].buffer]
                    self.pca_flag[i] = True
                    self.normalized[i] = True
                
                else:
                    pass

                for _ in range(self.update_per_time_step):
                    o, a, r, n, d = self.replay_buffer[i].sample(self.batch_size)
                    tensor = torch.cuda.FloatTensor if self.device.type == 'cuda' else torch.FloatTensor
                    o = tensor(o).to(self.device)
                    n = tensor(n).to(self.device)
                    a = tensor(a).to(self.device)
                    r = tensor(r).unsqueeze(1).to(self.device)
                    d = tensor(d).unsqueeze(1).to(self.device)

                    with torch.no_grad():
                        # Update Q-values. First, sample an action from the Gaussian policy/distribution for the current (next) observation and its associated log probability of occurrence.
                        new_next_actions, new_log_pi, _ = self.policy_net[i].sample(n)

                        # The updated Q-value is found by subtracting the logprob of the sampled action (proportional to the entropy) to the Q-values estimated by the target networks.
                        target_q_values = torch.min(
                            self.target_soft_q_net1[i](n, new_next_actions),
                            self.target_soft_q_net2[i](n, new_next_actions),
                        ) - self.alpha*new_log_pi
                        q_target = r + (1 - d)*self.discount*target_q_values

                    # Update Soft Q-Networks
                    q1_pred = self.soft_q_net1[i](o, a)
                    q2_pred = self.soft_q_net2[i](o, a)
                    q1_loss = self.soft_q_criterion(q1_pred, q_target)
                    q2_loss = self.soft_q_criterion(q2_pred, q_target)
                    self.soft_q_optimizer1[i].zero_grad()
                    q1_loss.backward()
                    self.soft_q_optimizer1[i].step()
                    self.soft_q_optimizer2[i].zero_grad()
                    q2_loss.backward()
                    self.soft_q_optimizer2[i].step()

                    # Update Policy
                    new_actions, log_pi, _ = self.policy_net[i].sample(o)
                    q_new_actions = torch.min(
                        self.soft_q_net1[i](o, new_actions),
                        self.soft_q_net2[i](o, new_actions)
                    )
                    policy_loss = (self.alpha*log_pi - q_new_actions).mean()
                    self.policy_optimizer[i].zero_grad()
                    policy_loss.backward()
                    self.policy_optimizer[i].step()

                    # Soft Updates
                    for target_param, param in zip(self.target_soft_q_net1[i].parameters(), self.soft_q_net1[i].parameters()):
                        target_param.data.copy_(target_param.data*(1.0 - self.tau) + param.data*self.tau)

                    for target_param, param in zip(self.target_soft_q_net2[i].parameters(), self.soft_q_net2[i].parameters()):
                        target_param.data.copy_(target_param.data*(1.0 - self.tau) + param.data*self.tau)

            else:
                pass

    def get_post_exploration_prediction(self, observations: List[List[float]], deterministic: bool) -> List[List[float]]:
        func = {
            True: self.get_post_exploration_prediction_with_information_sharing,
            False: self.get_post_exploration_prediction_without_information_sharing
        }[self.information_sharing]
        actions, coordination_variables = func(observations, deterministic)
        self.__coordination_variables_history[0] = deepcopy(self.__coordination_variables_history[1])
        self.__coordination_variables_history[1] = coordination_variables[0:]
        
        return actions

    def get_exploration_prediction(self, observations: List[List[float]]) -> List[List[float]]:
        func = {
            True: self.get_exploration_prediction_with_information_sharing,
            False: self.get_exploration_prediction_without_information_sharing
        }[self.information_sharing]
        actions, coordination_variables = func(observations)
        self.__coordination_variables_history[0] = deepcopy(self.__coordination_variables_history[1])
        self.__coordination_variables_history[1] = coordination_variables[0:]
        
        return actions

    def get_post_exploration_prediction_with_information_sharing(self, observations: List[List[float]], deterministic: bool) -> Tuple[List[List[float]], List[List[float]]]:
        agent_count = len(self.action_dimension)
        actions = [None for _ in range(agent_count)]
        action_order = list(range(agent_count))
        next_agent_ixs = [sorted(action_order)[action_order[(i + 1)%agent_count]] for i in range(agent_count)]
        coordination_variables = [[0.0, 0.0] for _ in range(agent_count)] 
        expected_demand = [0.0 for _ in range(agent_count)]
        total_demand = 0.0
        
        for i in range(self.iterations):
            capacity_dispatched = 0.0

            for c, n, o, o_ in zip(action_order, next_agent_ixs, observations, observations):
                o = self.get_encoded_observations(c, o)
                o = np.hstack(np.concatenate((o, coordination_variables[c])))
                o = self.get_normalized_observations(c, o)
                o = self.pca[c].transform(o.reshape(1,-1))[0]
                o = torch.FloatTensor(o).unsqueeze(0).to(self.device)
                result = self.policy_net[i].sample(o)
                a = result[2] if deterministic else result[0]
                a = list(a.detach().cpu().numpy()[0])
                actions[c] = a
                expected_demand[c] = self.predict_demand(c, o_, a)

                if i == self.iterations - 1 and c == action_order[-1]:
                    pass
                else:
                    total_demand += expected_demand[c] - expected_demand[n]
                    coordination_variables[n][0] = total_demand/self.total_coefficient

                coordination_variables[c][1] = capacity_dispatched
                capacity_dispatched += self.energy_size_coefficient[c]
        
        return actions, coordination_variables

    def get_post_exploration_prediction_without_information_sharing(self, observations: List[List[float]], deterministic: bool) -> Tuple[List[List[float]], List[List[float]]]:
        agent_count = len(self.action_dimension)
        actions = [None for _ in range(agent_count)]
        coordination_variables = [[0.0, 0.0] for _ in range(agent_count)]

        for i, o in enumerate(observations):
            o = self.get_encoded_observations(i, o)
            o = self.get_normalized_observations(i, o)
            o = self.pca[i].transform(o.reshape(1,-1))[0]
            o = torch.FloatTensor(o).unsqueeze(0).to(self.device)
            result = self.policy_net[i].sample(o)
            a = result[2] if deterministic else result[0]
            a = list(a.detach().cpu().numpy()[0])
            actions[i] = a
        
        return actions, coordination_variables

    def get_exploration_prediction_with_information_sharing(self, observations: List[List[float]]) -> Tuple[List[List[float]], List[List[float]]]:
        actions, coordination_variables = self.get_exploration_prediction_without_information_sharing(observations)
    
        if self.time_step > self.start_regression_time_step:
            agent_count = len(self.action_dimension)
            action_order = list(range(agent_count))
            nprs = np.random.RandomState(int(self.random_seed + self.time_step))
            nprs.shuffle(action_order)
            expected_demand = [self.predict_demand(i, o, a) for i, (o, a) in enumerate(zip(observations, actions))]
            coordination_variables = [[
                (sum(expected_demand) - expected_demand[i])/self.total_coefficient,
                sum([self.energy_size_coefficient[j] for j in action_order[0:action_order.index(i)]])
            ] for i in range(agent_count)]
        
        else:
            pass
        
        return actions, coordination_variables

    def get_exploration_prediction_without_information_sharing(self, observations: List[List[float]]) -> Tuple[List[List[float]], List[List[float]]]:
        actions = super().get_exploration_prediction(observations)
        coordination_variables = [[0.0, 0.0] for _ in range(len(self.action_dimension))]

        return actions, coordination_variables

    def predict_demand(self, index: int, observations: List[float], actions: List[float]) -> float:
        return self.state_estimator[index].predict(self.get_regression_variables(index, observations, actions).reshape(1, -1))[0]
    
    def get_regression_variables(self, index: int, observations: List[float], actions: List[float]) -> List[float]:
        return np.hstack(np.concatenate((self.get_encoded_regression_variables(index, observations), actions)))

    def get_encoded_regression_variables(self, index: int, observations: List[float]) -> List[float]:
        net_electricity_consumption_ix = self.observation_names[index].index('net_electricity_consumption')
        o = observations[0:]
        del o[net_electricity_consumption_ix]
        e = self.regression_encoders[index][0:]
        del e[net_electricity_consumption_ix]
        
        return np.array([j for j in np.hstack(e*np.array(o, dtype=float)) if j != None], dtype=float).tolist()

    def get_encoded_regression_targets(self, index: int, observations: List[float]) -> float:
        net_electricity_consumption_ix = self.observation_names[index].index('net_electricity_consumption')
        o = observations[net_electricity_consumption_ix]
        e = self.regression_encoders[index][net_electricity_consumption_ix]
        
        return e*o

    def set_pca(self):
        addition = self.__COORDINATION_VARIABLE_COUNT if self.information_sharing else 0
        
        for i, s in enumerate(self.observation_space):
            n_components = int((self.pca_compression)*(addition + len(self.get_encoded_observations(i, s.low))))
            self.pca[i] = PCA(n_components=n_components)

    def set_energy_coefficients(self):
        self.energy_size_coefficient = []
        self.total_coefficient = 0

        for b in self.building_metadata:
            coef = b['annual_dhw_demand_estimate']/.9 \
                + b['annual_cooling_demand_estimate']/3.5 \
                    + b['annual_heating_demand_estimate']/3.5 \
                        + b['annual_non_shiftable_load_estimate'] \
                            - b['annual_solar_generation_estimate']/6.0
            coef = max(0.3*(coef + b['annual_solar_generation_estimate']/6.0), coef)/8760
            self.energy_size_coefficient.append(coef)
            self.total_coefficient += coef

        self.energy_size_coefficient = [c/self.total_coefficient for c in self.energy_size_coefficient]

    def set_regression_encoders(self) -> List[List[Encoder]]:
        r"""Get observation value transformers/encoders for use in MARLISA agent internal regression model.

        The encoder classes are defined in the `preprocessing.py` module and include `PeriodicNormalization` for cyclic observations,
        `OnehotEncoding` for categorical obeservations, `RemoveFeature` for non-applicable observations given available storage systems and devices
        and `Normalize` for observations with known minimum and maximum boundaries.
        
        Returns
        -------
        encoders : List[Encoder]
            Encoder classes for observations ordered with respect to `active_observations`.
        """

        encoders = []
        remove_features = [
            'outdoor_dry_bulb_temperature', 'outdoor_dry_bulb_temperature_predicted_1',
            'outdoor_dry_bulb_temperature_predicted_2','outdoor_dry_bulb_temperature_predicted_3',
            'outdoor_relative_humidity', 'outdoor_relative_humidity_predicted_1',
            'outdoor_relative_humidity_predicted_2','outdoor_relative_humidity_predicted_3',
            'diffuse_solar_irradiance', 'diffuse_solar_irradiance_predicted_1',
            'diffuse_solar_irradiance_predicted_2', 'diffuse_solar_irradiance_predicted_3',
            'direct_solar_irradiance', 'direct_solar_irradiance_predicted_1',
            'direct_solar_irradiance_predicted_2', 'direct_solar_irradiance_predicted_3',
        ]

        for o, s in zip(self.observation_names, self.observation_space):
            e = []

            for i, n in enumerate(o):
                if n in ['month', 'hour']:
                    e.append(PeriodicNormalization(s.high[i]))
            
                elif n in remove_features:
                    e.append(RemoveFeature())
            
                else:
                    e.append(NoNormalization())

            encoders.append(e)

        return encoders

    def set_networks(self):
        internal_observation_count = self.__COORDINATION_VARIABLE_COUNT if self.information_sharing else 0
        return super().set_networks(internal_observation_count=internal_observation_count)

    def reset(self):
        super().reset()
        self.__coordination_variables_history = [
            [[0.0]*self.__COORDINATION_VARIABLE_COUNT for _ in self.action_dimension] for _ in range(2)
        ]

class MARLISARBC(MARLISA, SACRBC):
    r"""Uses :py:class:`citylearn.agents.rbc.RBC` to select action during exploration before using :py:class:`citylearn.agents.marlisa.MARLISA`.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    rbc: RBC
        :py:class:`citylearn.agents.rbc.RBC` or child class, used to select actions during exploration.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """
    
    def __init__(self, env: CityLearnEnv, rbc: RBC = None, **kwargs: Any):
        super().__init__(env=env, rbc=rbc, **kwargs)

    def get_exploration_prediction_without_information_sharing(self, observations: List[List[float]]) -> Tuple[List[List[float]], List[List[float]]]:
        _, coordination_variables = super().get_exploration_prediction_without_information_sharing(observations)
        actions = super(SACRBC, self).get_exploration_prediction(observations)

        return actions, coordination_variables


import logging
from typing import List, Mapping, Tuple
from gymnasium import spaces
import numpy as np
from citylearn.base import Environment, EpisodeTracker
from citylearn.energy_model import Battery
from citylearn.preprocessing import Normalize, PeriodicNormalization

ZERO_DIVISION_PLACEHOLDER = 0.000001
LOGGER = logging.getLogger()

class ElectricVehicle(Environment):

    def __init__(self, episode_tracker: EpisodeTracker,
                 battery: Battery = None, name: str = None, **kwargs):
        """
        Initialize the EVCar class.

        Parameters
        ----------
        battery : Battery
            An instance of the Battery class.
        name : str, optional
            Unique Electric_Vehicle name.

        Other Parameters
        ----------------
        **kwargs : dict
            Other keyword arguments used to initialize super class.
        """

        self.name = name

        super().__init__(
            seconds_per_time_step=kwargs.get('seconds_per_time_step'),
            random_seed=kwargs.get('random_seed'),
            episode_tracker=episode_tracker,
            time_step_ratio=battery.time_step_ratio
        )

        self.battery = battery
        self.__observation_epsilon = 0.0  # to avoid out of bound observations


    @property
    def name(self) -> str:
        """Unique building name."""

        return self.__name

    @property
    def battery(self) -> Battery:
        """Battery for Electric_Vehicle."""

        return self.__battery

    @name.setter
    def name(self, name: str):
        self.__name = name
    
    @battery.setter
    def battery(self, battery: Battery):
        if battery is None:
            raise AttributeError("Battery set to None")
        else:
            self.__battery = battery

    def next_time_step(self) -> Mapping[int, str]:        
        self.battery.next_time_step()
        super().next_time_step()

    def reset(self):
        """
        Reset the EVCar to its initial state.
        """

        super().reset()
        self.battery.reset()

    def observations(self) -> Mapping[str, float]:
        r"""Observations at current time step.

        Parameters
        ----------

        Returns
        -------
        observations
        """
        unwanted_keys = ["electric_vehicle_charger_state", "charger", "electric_vehicle_soc_arrival"]

        observations = {
            **{
                k.lstrip('_'): v[self.time_step]
                for k, v in vars(self).items()
                if isinstance(v, np.ndarray) and k.lstrip('_') not in unwanted_keys
                # Ensure filtering is done after stripping
            },
            'electric_vehicle_soc': self.battery.soc[self.time_step]
        }

        return observations

    # --- String / Object Representation Methods ---

    def __str__(self) -> str:
        """
        Return a text representation of the current state.
        """
        return str(self.as_dict())

    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the current state for use in rendering or logging.
        """
        return {
            'name': self.name,
            'Battery capacity': self.battery.capacity,  
            **self.observations() 
        }

    def render_simulation_end_data(self) -> dict:
        """
        Return a dictionary containing all simulation data across all time steps.
        The returned dictionary is structured with a general simulation name and, for each time step,
        a dictionary with the simulation data and battery data.

        Returns
        -------
        result : dict
            A JSON-like dictionary with the simulation name and per-time-step data.
        """
        # Determine the number of time steps.
        num_steps = self.episode_tracker.episode_time_steps

        # Gather simulation attributes (only those that are numpy arrays).
        # Build a list of dictionaries for each time step.
        time_steps = []
        for i in range(num_steps):
            step_data = {"time_step": i, "battery": {}}

            # Add battery data for this time step.
            soc_value = self.battery.soc[i]
            if isinstance(soc_value, np.generic):
                soc_value = soc_value.item()
            step_data["battery"] = {
                "soc": soc_value,
                "capacity": self.battery.capacity  # capacity is assumed constant over time.
            }

            time_steps.append(step_data)

        result = {
            "simulation_name": self.name if self.name else "ElectricVehicleSimulation",
            "data": time_steps
        }
        return result





